# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2014-2019 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Device specific settings and configuration"""


## Imports

from __future__ import division, print_function
import logging
import collections
import re

from lib.gettext import C_
from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import Pango

from lib.observable import event
import gui.application
import gui.mode

logger = logging.getLogger(__name__)

## Device prefs

# The per-device settings are stored in the prefs in a sub-dict whose
# string keys are formed from the device name and enough extra
# information to (hopefully) identify the device uniquely. Names are not
# unique, and IDs vary according to the order in which you plug devices
# in. So for now, our unique strings use a combination of the device's
# name, its source as presented by GDK, and the number of axes.

_PREFS_ROOT = "input.devices"
_PREFS_DEVICE_SUBKEY_FMT = "{name}:{source}:{num_axes}"


## Device type strings

_DEVICE_TYPE_STRING = {
    Gdk.InputSource.CURSOR: C_(
        "prefs: device's type label",
        "Cursor/puck",
    ),
    Gdk.InputSource.ERASER: C_(
        "prefs: device's type label",
        "Eraser",
    ),
    Gdk.InputSource.KEYBOARD: C_(
        "prefs: device's type label",
        "Keyboard",
    ),
    Gdk.InputSource.MOUSE: C_(
        "prefs: device's type label",
        "Mouse",
    ),
    Gdk.InputSource.PEN: C_(
        "prefs: device's type label",
        "Pen",
    ),
    Gdk.InputSource.TOUCHPAD: C_(
        "prefs: device's type label",
        "Touchpad",
    ),
    Gdk.InputSource.TOUCHSCREEN: C_(
        "prefs: device's type label",
        "Touchscreen",
    ),
}


## Settings consts and classes


class AllowedUsage:
    """Consts describing how a device may interact with the canvas"""

    ANY = "any"  #: Device can be used for any tasks.
    NOPAINT = "nopaint"  #: No direct painting, but can manipulate objects.
    NAVONLY = "navonly"  #: Device can only be used for navigation.
    IGNORED = "ignored"  #: Device cannot interact with the canvas at all.

    VALUES = (ANY, IGNORED, NOPAINT, NAVONLY)
    DISPLAY_STRING = {
        IGNORED: C_(
            "device settings: allowed usage",
            u"Ignore",
        ),
        ANY: C_(
            "device settings: allowed usage",
            u"Any Task",
        ),
        NOPAINT: C_(
            "device settings: allowed usage",
            u"Non-painting tasks",
        ),
        NAVONLY: C_(
            "device settings: allowed usage",
            u"Navigation only",
        ),
    }
    BEHAVIOR_MASK = {
        ANY: gui.mode.Behavior.ALL,
        IGNORED: gui.mode.Behavior.NONE,
        NOPAINT: gui.mode.Behavior.NON_PAINTING,
        NAVONLY: gui.mode.Behavior.CHANGE_VIEW,
    }


class ScrollAction:
    """Consts describing how a device's scroll events should be used.

    The user can assign one of these values to a device to configure
    whether they'd prefer panning or scrolling for unmodified scroll
    events. This setting can be queried via the device monitor.

    """

    ZOOM = "zoom"  #: Alter the canvas scaling
    PAN = "pan"   #: Pan across the canvas

    VALUES = (ZOOM, PAN)
    DISPLAY_STRING = {
        ZOOM: C_("device settings: unmodified scroll action", u"Zoom"),
        PAN: C_("device settings: unmodified scroll action", u"Pan"),
    }


class Settings (object):
    """A device's settings"""

    DEFAULT_USAGE = AllowedUsage.VALUES[0]
    DEFAULT_SCROLL = ScrollAction.VALUES[0]

    def __init__(self, prefs, usage=DEFAULT_USAGE, scroll=DEFAULT_SCROLL):
        super(Settings, self).__init__()
        self._usage = self.DEFAULT_USAGE
        self._update_usage_mask()
        self._scroll = self.DEFAULT_SCROLL
        self._prefs = prefs
        self._load_from_prefs()

    @property
    def usage(self):
        return self._usage

    @usage.setter
    def usage(self, value):
        if value not in AllowedUsage.VALUES:
            raise ValueError("Unrecognized usage value")
        self._usage = value
        self._update_usage_mask()
        self._save_to_prefs()

    @property
    def usage_mask(self):
        return self._usage_mask

    @property
    def scroll(self):
        return self._scroll

    @scroll.setter
    def scroll(self, value):
        if value not in ScrollAction.VALUES:
            raise ValueError("Unrecognized scroll value")
        self._scroll = value
        self._save_to_prefs()

    def _load_from_prefs(self):
        usage = self._prefs.get("usage", self.DEFAULT_USAGE)
        if usage not in AllowedUsage.VALUES:
            usage = self.DEFAULT_USAGE
        self._usage = usage
        scroll = self._prefs.get("scroll", self.DEFAULT_SCROLL)
        if scroll not in ScrollAction.VALUES:
            scroll = self.DEFAULT_SCROLL
        self._scroll = scroll
        self._update_usage_mask()

    def _save_to_prefs(self):
        self._prefs.update({
            "usage": self._usage,
            "scroll": self._scroll,
        })

    def _update_usage_mask(self):
        self._usage_mask = AllowedUsage.BEHAVIOR_MASK[self._usage]


## Main class defs


class Monitor (object):
    """Monitors device use & plugging, and manages their configuration

    An instance resides in the main application. It is responsible for
    monitoring known devices, determining their characteristics, and
    storing their settings. Per-device settings are stored in the main
    application preferences.

    """

    def __init__(self, app):
        """Initializes, assigning initial input device uses

        :param app: the owning Application instance.
        :type app: gui.application.Application
        """
        super(Monitor, self).__init__()
        self._app = app
        if app is not None:
            self._prefs = app.preferences
        else:
            self._prefs = {}
        if _PREFS_ROOT not in self._prefs:
            self._prefs[_PREFS_ROOT] = {}

        # Transient device information
        self._device_settings = collections.OrderedDict()  # {dev: settings}
        self._last_event_device = None
        self._last_pen_device = None

        disp = Gdk.Display.get_default()
        mgr = disp.get_device_manager()
        mgr.connect("device-added", self._device_added_cb)
        mgr.connect("device-removed", self._device_removed_cb)
        self._device_manager = mgr

        for physical_device in mgr.list_devices(Gdk.DeviceType.SLAVE):
            self._init_device_settings(physical_device)

    ## Devices list

    def get_device_settings(self, device):
        """Gets the settings for a device

        :param Gdk.Device device: a physical ("slave") device
        :returns: A settings object which can be manipulated, or None
        :rtype: Settings

        Changes to the returned object made via its API are saved to the
        user preferences immediately.

        If the device is a keyboard, or is otherwise unsuitable as a
        pointing device, None is returned instead. The caller needs to
        check this case.

        """
        self._init_device_settings(device)
        return self._device_settings.get(device)

    def _init_device_settings(self, device):
        """Ensures that the device settings are loaded for a device"""
        source = device.get_source()
        if source == Gdk.InputSource.KEYBOARD:
            return
        num_axes = device.get_n_axes()
        if num_axes < 2:
            return
        settings = self._device_settings.get(device)
        if not settings:
            try:
                vendor_id = device.get_vendor_id()
                product_id = device.get_product_id()
            except AttributeError:
                # New in GDK 3.16
                vendor_id = "?"
                product_id = "?"
            logger.info(
                "New device %r"
                " (%s, axes:%d, class=%s, vendor=%r, product=%r)",
                device.get_name(),
                source.value_name,
                num_axes,
                device.__class__.__name__,
                vendor_id,
                product_id,
            )
            dev_prefs_key = _device_prefs_key(device)
            dev_prefs = self._prefs[_PREFS_ROOT].setdefault(dev_prefs_key, {})
            settings = Settings(dev_prefs)
            self._device_settings[device] = settings
            self.devices_updated()
        assert settings is not None

    def _device_added_cb(self, mgr, device):
        """Informs that a device has been plugged in"""
        logger.debug("device-added %r", device.get_name())
        self._init_device_settings(device)

    def _device_removed_cb(self, mgr, device):
        """Informs that a device has been unplugged"""
        logger.debug("device-removed %r", device.get_name())
        self._device_settings.pop(device, None)
        self.devices_updated()

    @event
    def devices_updated(self):
        """Event: the devices list was changed"""

    def get_devices(self):
        """Yields devices and their settings, for UI stuff

        :rtype: iterator
        :returns: ultimately a sequence of (Gdk.Device, Settings) pairs

        """
        for device, settings in self._device_settings.items():
            yield (device, settings)

    ## Current device

    @event
    def current_device_changed(self, old_device, new_device):
        """Event: the current device has changed

        :param Gdk.Device old_device: Previous device used
        :param Gdk.Device new_device: New device used
        """

    def device_used(self, device):
        """Informs about a device being used, for use by controllers

        :param Gdk.Device device: the device being used
        :returns: whether the device changed

        If the device has changed, this method then notifies interested
        parties via the device_changed observable @event.

        This method returns True if the device was the same as the previous
        device, and False if it has changed.
        """
        if not self.get_device_settings(device):
            return False
        if device == self._last_event_device:
            return True
        self.current_device_changed(self._last_event_device, device)
        old_device = self._last_event_device
        new_device = device
        self._last_event_device = device

        # small problem with this code: it doesn't work well with brushes that
        # have (eraser not in [1.0, 0.0])

        new_device.name = new_device.props.name
        new_device.source = new_device.props.input_source

        logger.debug(
            "Device change: name=%r source=%s",
            new_device.name, new_device.source.value_name,
        )

        # When editing brush settings, it is often more convenient to use the
        # mouse. Because of this, we don't restore brushsettings when switching
        # to/from the mouse. We act as if the mouse was identical to the last
        # active pen device.

        if (new_device.source == Gdk.InputSource.MOUSE and
                self._last_pen_device):
            new_device = self._last_pen_device
        if new_device.source == Gdk.InputSource.PEN:
            self._last_pen_device = new_device
        if (old_device and old_device.source == Gdk.InputSource.MOUSE and
                self._last_pen_device):
            old_device = self._last_pen_device

        bm = self._app.brushmanager
        if old_device:
            # Clone for saving
            old_brush = bm.clone_selected_brush(name=None)
            bm.store_brush_for_device(old_device.name, old_brush)

        if new_device.source == Gdk.InputSource.MOUSE:
            # Avoid fouling up unrelated devbrushes at stroke end
            self._prefs.pop('devbrush.last_used', None)
        else:
            # Select the brush and update the UI.
            # Use a sane default if there's nothing associated
            # with the device yet.
            brush = bm.fetch_brush_for_device(new_device.name)
            if brush is None:
                if device_is_eraser(new_device):
                    brush = bm.get_default_eraser()
                else:
                    brush = bm.get_default_brush()
            self._prefs['devbrush.last_used'] = new_device.name
            bm.select_brush(brush)


class SettingsEditor (Gtk.Grid):
    """Per-device settings editor"""

    ## Class consts

    _USAGE_CONFIG_COL = 0
    _USAGE_STRING_COL = 1
    _SCROLL_CONFIG_COL = 0
    _SCROLL_STRING_COL = 1

    __gtype_name__ = "MyPaintDeviceSettingsEditor"

    ## Initialization

    def __init__(self, monitor=None):
        """Initialize

        :param Monitor monitor: monitor instance (for testing)

        By default, the central app's `device_monitor` is used to permit
        parameterless construction.
        """
        super(SettingsEditor, self).__init__()
        if monitor is None:
            app = gui.application.get_app()
            monitor = app.device_monitor
        self._monitor = monitor

        self._devices_store = Gtk.ListStore(object)
        self._devices_view = Gtk.TreeView(model=self._devices_store)

        col = Gtk.TreeViewColumn(C_(
            "prefs: devices table: column header",
            # TRANSLATORS: Column's data is the device's name
            "Device",
        ))
        col.set_min_width(200)
        col.set_expand(True)
        col.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self._devices_view.append_column(col)
        cell = Gtk.CellRendererText()
        cell.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        col.pack_start(cell, True)
        col.set_cell_data_func(cell, self._device_name_datafunc)

        col = Gtk.TreeViewColumn(C_(
            "prefs: devices table: column header",
            # TRANSLATORS: Column's data is the number of axes (an integer)
            "Axes",
        ))
        col.set_min_width(30)
        col.set_resizable(True)
        col.set_expand(False)
        col.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self._devices_view.append_column(col)
        cell = Gtk.CellRendererText()
        col.pack_start(cell, True)
        col.set_cell_data_func(cell, self._device_axes_datafunc)

        col = Gtk.TreeViewColumn(C_(
            "prefs: devices table: column header",
            # TRANSLATORS: Column shows type labels ("Touchscreen", "Pen" etc.)
            "Type",
        ))
        col.set_min_width(120)
        col.set_resizable(True)
        col.set_expand(False)
        col.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self._devices_view.append_column(col)
        cell = Gtk.CellRendererText()
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        col.pack_start(cell, True)
        col.set_cell_data_func(cell, self._device_type_datafunc)

        # Usage config value => string store (dropdowns)
        store = Gtk.ListStore(str, str)
        for conf_val in AllowedUsage.VALUES:
            string = AllowedUsage.DISPLAY_STRING[conf_val]
            store.append([conf_val, string])
        self._usage_store = store

        col = Gtk.TreeViewColumn(C_(
            "prefs: devices table: column header",
            # TRANSLATORS: Column's data is a dropdown allowing the allowed
            # TRANSLATORS: tasks for the row's device to be configured.
            u"Use for…",
        ))
        col.set_min_width(100)
        col.set_resizable(True)
        col.set_expand(False)
        self._devices_view.append_column(col)

        cell = Gtk.CellRendererCombo()
        cell.set_property("model", self._usage_store)
        cell.set_property("text-column", self._USAGE_STRING_COL)
        cell.set_property("mode", Gtk.CellRendererMode.EDITABLE)
        cell.set_property("editable", True)
        cell.set_property("has-entry", False)
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        cell.connect("changed", self._usage_cell_changed_cb)
        col.pack_start(cell, True)
        col.set_cell_data_func(cell, self._device_usage_datafunc)

        # Scroll action config value => string store (dropdowns)
        store = Gtk.ListStore(str, str)
        for conf_val in ScrollAction.VALUES:
            string = ScrollAction.DISPLAY_STRING[conf_val]
            store.append([conf_val, string])
        self._scroll_store = store

        col = Gtk.TreeViewColumn(C_(
            "prefs: devices table: column header",
            # TRANSLATORS: Column's data is a dropdown for how the device's
            # TRANSLATORS: scroll wheel or scroll-gesture events are to be
            # TRANSLATORS: interpreted normally.
            u"Scroll…",
        ))
        col.set_min_width(100)
        col.set_resizable(True)
        col.set_expand(False)
        self._devices_view.append_column(col)

        cell = Gtk.CellRendererCombo()
        cell.set_property("model", self._scroll_store)
        cell.set_property("text-column", self._USAGE_STRING_COL)
        cell.set_property("mode", Gtk.CellRendererMode.EDITABLE)
        cell.set_property("editable", True)
        cell.set_property("has-entry", False)
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        cell.connect("changed", self._scroll_cell_changed_cb)
        col.pack_start(cell, True)
        col.set_cell_data_func(cell, self._device_scroll_datafunc)

        # Pretty borders
        view_scroll = Gtk.ScrolledWindow()
        view_scroll.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        pol = Gtk.PolicyType.AUTOMATIC
        view_scroll.set_policy(pol, pol)
        view_scroll.add(self._devices_view)
        view_scroll.set_hexpand(True)
        view_scroll.set_vexpand(True)
        self.attach(view_scroll, 0, 0, 1, 1)

        self._update_devices_store()
        self._monitor.devices_updated += self._update_devices_store

    ## Display and sort funcs

    def _device_name_datafunc(self, column, cell, model, iter_, *data):
        device = model.get_value(iter_, 0)
        cell.set_property("text", device.get_name())

    def _device_axes_datafunc(self, column, cell, model, iter_, *data):
        device = model.get_value(iter_, 0)
        n_axes = device.get_n_axes()
        cell.set_property("text", "%d" % (n_axes,))

    def _device_type_datafunc(self, column, cell, model, iter_, *data):
        device = model.get_value(iter_, 0)
        source = device.get_source()
        text = _DEVICE_TYPE_STRING.get(source, source.value_nick)
        cell.set_property("text", text)

    def _device_usage_datafunc(self, column, cell, model, iter_, *data):
        device = model.get_value(iter_, 0)
        settings = self._monitor.get_device_settings(device)
        if not settings:
            return
        text = AllowedUsage.DISPLAY_STRING[settings.usage]
        cell.set_property("text", text)

    def _device_scroll_datafunc(self, column, cell, model, iter_, *data):
        device = model.get_value(iter_, 0)
        settings = self._monitor.get_device_settings(device)
        if not settings:
            return
        text = ScrollAction.DISPLAY_STRING[settings.scroll]
        cell.set_property("text", text)

    ## Updates

    def _usage_cell_changed_cb(self, combo, device_path_str,
                               usage_iter, *etc):
        config = self._usage_store.get_value(
            usage_iter,
            self._USAGE_CONFIG_COL,
        )
        device_iter = self._devices_store.get_iter(device_path_str)
        device = self._devices_store.get_value(device_iter, 0)
        settings = self._monitor.get_device_settings(device)
        if not settings:
            return
        settings.usage = config
        self._devices_view.columns_autosize()

    def _scroll_cell_changed_cb(self, conf_combo, device_path_str,
                                conf_iter, *etc):
        conf_store = self._scroll_store
        conf_col = self._SCROLL_CONFIG_COL
        conf_value = conf_store.get_value(conf_iter, conf_col)
        device_store = self._devices_store
        device_iter = device_store.get_iter(device_path_str)
        device = device_store.get_value(device_iter, 0)
        settings = self._monitor.get_device_settings(device)
        if not settings:
            return
        settings.scroll = conf_value
        self._devices_view.columns_autosize()

    def _update_devices_store(self, *_ignored):
        """Repopulates the displayed list"""
        updated_list = list(self._monitor.get_devices())
        updated_list_map = dict(updated_list)
        paths_for_removal = []
        devices_retained = set()
        for row in self._devices_store:
            device, = row
            if device not in updated_list_map:
                paths_for_removal.append(row.path)
                continue
            devices_retained.add(device)
        for device, config in updated_list:
            if device in devices_retained:
                continue
            self._devices_store.append([device])
        for unwanted_row_path in reversed(paths_for_removal):
            unwanted_row_iter = self._devices_store.get_iter(unwanted_row_path)
            self._devices_store.remove(unwanted_row_iter)
        self._devices_view.queue_draw()


## Helper funcs


def _device_prefs_key(device):
    """Returns the subkey to use in the app prefs for a device"""
    source = device.get_source()
    name = device.get_name()
    n_axes = device.get_n_axes()
    return u"%s:%s:%d" % (name, source.value_nick, n_axes)


def device_is_eraser(device):
    """Tests whether a device appears to be an eraser"""
    if device is None:
        return False
    if device.get_source() == Gdk.InputSource.ERASER:
        return True
    if re.search(r'\<eraser\>', device.get_name(), re.I):
        return True
    return False


## Testing


def _test():
    """Interactive UI testing for SettingsEditor and Monitor"""
    logging.basicConfig(level=logging.DEBUG)
    win = Gtk.Window()
    win.set_title("gui.device.SettingsEditor")
    win.set_default_size(500, 400)
    win.connect("destroy", Gtk.main_quit)
    monitor = Monitor(app=None)
    editor = SettingsEditor(monitor)
    win.add(editor)
    win.show_all()
    Gtk.main()
    print(monitor._prefs)


if __name__ == '__main__':
    _test()



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2015 by Andrew Chadwick <a.t.chadwick@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""UI behaviour for picking things from the canvas.

The grab and button behaviour objects work like MVP presenters
with a rather wide scope.

"""

## Imports
from __future__ import division, print_function

from gui.tileddrawwidget import TiledDrawWidget
from gui.document import Document
from lib.gettext import C_
import gui.cursor

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib

import abc
import logging
logger = logging.getLogger(__name__)


## Class definitions

class PickingGrabPresenter (object):
    """Picking something via a grab (abstract base, MVP presenter)

    This presenter mediates between passive GTK view widgets
    accessed via the central app,
    and a model consisting of some drawing state within the application.
    When activated, it establishes a pointer grab and a keyboard grab,
    updates the thing being grabbed zero or more times,
    then exits making sure that the grab is cleaned up correctly.

    """

    ## Class configuration

    __metaclass__ = abc.ABCMeta

    _GRAB_MASK = (Gdk.EventMask.BUTTON_RELEASE_MASK
                  | Gdk.EventMask.BUTTON_PRESS_MASK
                  | Gdk.EventMask.BUTTON_MOTION_MASK)

    ## Initialization

    def __init__(self):
        """Basic initialization."""
        super(PickingGrabPresenter, self).__init__()
        self._app = None
        self._statusbar_info_cache = None
        self._grab_button_num = None
        self._grabbed_pointer_dev = None
        self._grabbed_keyboard_dev = None
        self._grab_event_handler_ids = None
        self._delayed_picking_update_id = None

    @property
    def app(self):
        """The coordinating app object."""
        # FIXME: The view (statusbar, grab owner widget) is accessed
        # FIXME: through this, which may be a problem in the long term.
        # FIXME: There's a need to set up event masks before starting
        # FIXME: the grab, and this may make _start_grab() more fragile.
        # Ref: https://github.com/mypaint/mypaint/issues/324
        return self._app

    @app.setter
    def app(self, app):
        self._app = app
        self._statusbar_info_cache = None

    ## Internals

    @property
    def _grab_owner(self):
        """The view widget owning the grab."""
        return self.app.drawWindow

    @property
    def _statusbar_info(self):
        """The view widget and context for displaying status msgs."""
        if not self._statusbar_info_cache:
            statusbar = self.app.statusbar
            cid = statusbar.get_context_id("picker-button")
            self._statusbar_info_cache = (statusbar, cid)
        return self._statusbar_info_cache

    def _hide_status_message(self):
        """Remove all statusbar messages coming from this class"""
        statusbar, cid = self._statusbar_info
        statusbar.remove_all(cid)

    def _show_status_message(self):
        """Display a status message via the view."""
        statusbar, cid = self._statusbar_info
        statusbar.push(cid, self.picking_status_text)

    ## Activation

    def activate_from_button_event(self, event):
        """Activate during handling of a GdkEventButton (press/release)

        If the event is a button press, then the grab will start
        immediately, begin updating immediately, and will terminate by
        the release of the initiating button.

        If the event is a button release, then the grab start will be
        deferred to start in an idle handler. When the grab starts, it
        won't begin updating until the user clicks button 1 (and only
        button 1), and it will only be terminated with a button1
        release. This covers the case of events delivered to "clicked"
        signal handlers

        """
        if event.type == Gdk.EventType.BUTTON_PRESS:
            logger.debug("Starting picking grab")
            has_button_info, button_num = event.get_button()
            if not has_button_info:
                return
            self._start_grab(event.device, event.time, button_num)
        elif event.type == Gdk.EventType.BUTTON_RELEASE:
            logger.debug("Queueing picking grab")
            GLib.idle_add(
                self._start_grab,
                event.device,
                event.time,
                None,
            )

    ## Required interface for subclasses

    @abc.abstractproperty
    def picking_cursor(self):
        """The cursor to use while picking.

        :returns: The cursor to use during the picking grab.
        :rtype: Gdk.Cursor

        This abstract property must be overridden with an implementation
        giving an appropriate cursor to display during the picking grab.

        """

    @abc.abstractproperty
    def picking_status_text(self):
        """The statusbar text to use during the grab."""

    @abc.abstractmethod
    def picking_update(self, device, x_root, y_root):
        """Update whatever's being picked during & after picking.

        :param Gdk.Device device: Pointer device currently grabbed
        :param int x_root: Absolute screen X coordinate
        :param int y_root: Absolute screen Y coordinate

        This abstract method must be overridden with an implementation
        which updates the model object being picked.
        It is always called at the end of the picking grab
        when button1 is released,
        and may be called several times during the grab
        while button1 is held.

        See gui.tileddrawwidget.TiledDrawWidget.get_tdw_under_device()
        for details of how to get canvas widgets
        and their related document models and controllers.

        """

    ## Internals

    def _start_grab(self, device, time, inibutton):
        """Start the pointer grab, and enter the picking state.

        :param Gdk.Device device: Initiating pointer device.
        :param int time: The grab start timestamp.
        :param int inibutton: Initiating pointer button.

        The associated keyboard device is grabbed too.
        This method assumes that inibutton is currently held. The grab
        terminates when inibutton is released.

        """
        logger.debug("Starting picking grab...")

        # The device to grab must be a virtual device,
        # because we need to grab its associated virtual keyboard too.
        # We don't grab physical devices directly.
        if device.get_device_type() == Gdk.DeviceType.SLAVE:
            device = device.get_associated_device()
        elif device.get_device_type() == Gdk.DeviceType.FLOATING:
            logger.warning(
                "Cannot start grab on floating device %r",
                device.get_name(),
            )
            return
        assert device.get_device_type() == Gdk.DeviceType.MASTER

        # Find the keyboard paired to this pointer.
        assert device.get_source() != Gdk.InputSource.KEYBOARD
        keyboard_device = device.get_associated_device()  # again! top API!
        assert keyboard_device.get_device_type() == Gdk.DeviceType.MASTER
        assert keyboard_device.get_source() == Gdk.InputSource.KEYBOARD

        # Internal state checks
        assert not self._grabbed_pointer_dev
        assert not self._grab_button_num
        assert self._grab_event_handler_ids is None

        # Validate the widget we're expected to grab.
        owner = self._grab_owner
        assert owner.get_has_window()
        window = owner.get_window()
        assert window is not None

        # Ensure that it'll receive termination events.
        owner.add_events(self._GRAB_MASK)
        assert (int(owner.get_events() & self._GRAB_MASK) == int(self._GRAB_MASK)), \
            "Grab owner's events must match %r" % (self._GRAB_MASK,)

        # There should be no message in the statusbar from this Grab,
        # but clear it out anyway.
        self._hide_status_message()

        # Grab item, pointer first
        result = device.grab(
            window = window,
            grab_ownership = Gdk.GrabOwnership.APPLICATION,
            owner_events = False,
            event_mask = self._GRAB_MASK,
            cursor = self.picking_cursor,
            time_ = time,
        )
        if result != Gdk.GrabStatus.SUCCESS:
            logger.error(
                "Failed to create pointer grab on %r. "
                "Result: %r.",
                device.get_name(),
                result,
            )
            device.ungrab(time)
            return False  # don't requeue

        # Need to grab the keyboard too, since Mypaint uses hotkeys.
        keyboard_mask = Gdk.EventMask.KEY_PRESS_MASK \
            | Gdk.EventMask.KEY_RELEASE_MASK
        result = keyboard_device.grab(
            window = window,
            grab_ownership = Gdk.GrabOwnership.APPLICATION,
            owner_events = False,
            event_mask = keyboard_mask,
            cursor = self.picking_cursor,
            time_ = time,
        )
        if result != Gdk.GrabStatus.SUCCESS:
            logger.error(
                "Failed to create grab on keyboard associated with %r. "
                "Result: %r",
                device.get_name(),
                result,
            )
            device.ungrab(time)
            keyboard_device.ungrab(time)
            return False  # don't requeue

        # Grab is established
        self._grabbed_pointer_dev = device
        self._grabbed_keyboard_dev = keyboard_device
        logger.debug(
            "Grabs established on pointer %r and keyboard %r",
            device.get_name(),
            keyboard_device.get_name(),
        )

        # Tell the user how to work the thing.
        self._show_status_message()

        # Establish temporary event handlers during the grab.
        # These are responsible for ending the grab state.
        handlers = {
            "button-release-event": self._in_grab_button_release_cb,
            "motion-notify-event": self._in_grab_motion_cb,
            "grab-broken-event": self._in_grab_grab_broken_cb,
        }
        if not inibutton:
            handlers["button-press-event"] = self._in_grab_button_press_cb
        else:
            self._grab_button_num = inibutton
        handler_ids = []
        for signame, handler_cb in handlers.items():
            hid = owner.connect(signame, handler_cb)
            handler_ids.append(hid)
            logger.debug("Added handler for %r: hid=%d", signame, hid)
        self._grab_event_handler_ids = handler_ids

        return False   # don't requeue

    def _in_grab_button_press_cb(self, widget, event):
        assert self._grab_button_num is None
        if event.type != Gdk.EventType.BUTTON_PRESS:
            return False
        if not self._check_event_devices_still_grabbed(event):
            return
        has_button_info, button_num = event.get_button()
        if not has_button_info:
            return False
        if event.device is not self._grabbed_pointer_dev:
            return False
        self._grab_button_num = button_num
        return True

    def _in_grab_button_release_cb(self, widget, event):
        assert self._grab_button_num is not None
        if event.type != Gdk.EventType.BUTTON_RELEASE:
            return False
        if not self._check_event_devices_still_grabbed(event):
            return
        has_button_info, button_num = event.get_button()
        if not has_button_info:
            return False
        if button_num != self._grab_button_num:
            return False
        if event.device is not self._grabbed_pointer_dev:
            return False
        self._end_grab(event)
        assert self._grab_button_num is None
        return True

    def _in_grab_motion_cb(self, widget, event):
        assert self._grabbed_pointer_dev is not None
        if not self._check_event_devices_still_grabbed(event):
            return True
        if event.device is not self._grabbed_pointer_dev:
            return False
        if not self._grab_button_num:
            return False
        # Due to a performance issue, picking can take more time
        # than we have between two motion events (about 8ms).
        if self._delayed_picking_update_id:
            GLib.source_remove(self._delayed_picking_update_id)
        self._delayed_picking_update_id = GLib.idle_add(
            self._delayed_picking_update_cb,
            event.device,
            event.x_root,
            event.y_root,
        )
        return True

    def _in_grab_grab_broken_cb(self, widget, event):
        logger.debug("Grab broken, cleaning up.")
        self._ungrab_grabbed_devices(time=event.time)
        return False

    def _end_grab(self, event):
        """Finishes the picking grab normally."""
        if not self._check_event_devices_still_grabbed(event):
            return
        device = event.device
        try:
            self.picking_update(device, event.x_root, event.y_root)
        finally:
            self._ungrab_grabbed_devices(time=event.time)

    def _check_event_devices_still_grabbed(self, event):
        """Abandon picking if devices aren't still grabbed.

        This can happen if the escape key is pressed during the grab -
        the gui.keyboard handler is still invoked in the normal way,
        and Escape just does an ungrab.

        """
        cleanup_needed = False
        for dev in (self._grabbed_pointer_dev, self._grabbed_keyboard_dev):
            if not dev:
                cleanup_needed = True
                continue
            display = dev.get_display()
            if not display.device_is_grabbed(dev):
                logger.debug(
                    "Device %r is no longer grabbed: will clean up",
                    dev.get_name(),
                )
                cleanup_needed = True
        if cleanup_needed:
            self._ungrab_grabbed_devices(time=event.time)
        return not cleanup_needed

    def _ungrab_grabbed_devices(self, time=Gdk.CURRENT_TIME):
        """Ungrabs devices thought to be grabbed, and cleans up."""
        for dev in (self._grabbed_pointer_dev, self._grabbed_keyboard_dev):
            if not dev:
                continue
            logger.debug("Ungrabbing device %r", dev.get_name())
            dev.ungrab(time)
        # Unhook potential grab leave handlers
        # but only if the pick succeeded.
        if self._grab_event_handler_ids:
            for hid in self._grab_event_handler_ids:
                owner = self._grab_owner
                owner.disconnect(hid)
        self._grab_event_handler_ids = None
        # Update state (prevents the idler updating a 2nd time)
        self._grabbed_pointer_dev = None
        self._grabbed_keyboard_dev = None
        self._grab_button_num = None
        self._hide_status_message()

    def _delayed_picking_update_cb(self, ptrdev, x_root, y_root):
        """Delayed picking updates during grab.

        Some picking operations can be CPU-intensive, so this is called
        by an idle handler. If the user clicks and releases immediately,
        this never gets called, so a final call to picking_update() is
        made separately after the grab finishes.

        See: picking_update().

        """
        try:
            if ptrdev is self._grabbed_pointer_dev:
                self.picking_update(ptrdev, x_root, y_root)
        except:
            logger.exception("Exception in picking idler")
            # HMM: if it's not logged here, it won't be recorded...
        finally:
            self._delayed_picking_update_id = None
            return False

class ContextPickingGrabPresenter (PickingGrabPresenter):
    """Context picking behaviour (concrete MVP presenter)"""

    @property
    def picking_cursor(self):
        """The cursor to use while picking"""
        return self.app.cursors.get_icon_cursor(
            icon_name = "mypaint-brush-tip-symbolic",
            cursor_name = gui.cursor.Name.CROSSHAIR_OPEN_PRECISE,
        )

    @property
    def picking_status_text(self):
        """The statusbar text to use during the grab."""
        return C_(
            "context picker: statusbar text during grab",
            u"Pick brushstroke settings, stroke color, and layer…",
        )

    def picking_update(self, device, x_root, y_root):
        """Update brush and layer during & after picking."""
        # Can only pick from TDWs
        tdw, x, y = TiledDrawWidget.get_tdw_under_device(device)
        if tdw is None:
            return
        # Determine which document controller owns that tdw
        doc = None
        for d in Document.get_instances():
            if tdw is d.tdw:
                doc = d
                break
        if doc is None:
            return
        # Get that controller to do the pick.
        # Arguably this should be direct to the model.
        x, y = tdw.display_to_model(x, y)
        doc.pick_context(x, y)

class ColorPickingGrabPresenter (PickingGrabPresenter):
    """Color picking behaviour (concrete MVP presenter)"""

    @property
    def picking_cursor(self):
        """The cursor to use while picking"""
        return self.app.cursors.get_icon_cursor(
            icon_name = "mypaint-colors-symbolic",
            cursor_name = gui.cursor.Name.PICKER,
        )

    @property
    def picking_status_text(self):
        """The statusbar text to use during the grab."""
        return C_(
            "color picker: statusbar text during grab",
            u"Pick color…",
        )

    def picking_update(self, device, x_root, y_root):
        """Update brush and layer during & after picking."""
        tdw, x, y = TiledDrawWidget.get_tdw_under_device(device)
        if tdw is None:
            return
        color = tdw.pick_color(x, y)
        cm = self.app.brush_color_manager
        cm.set_color(color)


class ButtonPresenter (object):
    """Picking behaviour for a button (MVP presenter)

    This presenter mediates between a passive view consisting of a
    button, and a peer PickingGrabPresenter instance which does the
    actual work after the button is clicked.

    """

    ## Initialization

    def __init__(self):
        """Initialize."""
        super(ButtonPresenter, self).__init__()
        self._evbox = None
        self._button = None
        self._grab = None

    def set_picking_grab(self, grab):
        self._grab = grab

    def set_button(self, button):
        """Connect view button.

        :param Gtk.Button button: the initiator button

        """
        button.connect("clicked", self._clicked_cb)
        self._button = button

    ## Event handling

    def _clicked_cb(self, button):
        """Handle click events on the initiator button."""
        event = Gtk.get_current_event()
        assert event is not None
        assert event.type == Gdk.EventType.BUTTON_RELEASE, (
            "The docs lie! Current event's type is %r." % (event.type,),
        )
        self._grab.activate_from_button_event(event)



# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Component sliders for power users.
"""

from __future__ import division, print_function

from lib.gibindings import Gtk

from lib.color import RGBColor
from lib.color import HSVColor
from lib.color import HCYColor
from .bases import IconRenderable
from .adjbases import ColorAdjuster
from .adjbases import SliderColorAdjuster
from .combined import CombinedAdjusterPage
from lib.gettext import C_


class ComponentSlidersAdjusterPage (CombinedAdjusterPage, IconRenderable):
    """Component sliders for precise adjustment: page for `CombinedAdjuster`.
    """

    def __init__(self):
        CombinedAdjusterPage.__init__(self)
        grid = Gtk.Grid()
        grid.set_size_request(150, -1)
        grid.set_row_spacing(6)
        grid.set_column_spacing(0)
        grid.set_border_width(6)
        self._sliders = []   #: List of slider widgets.
        grid.set_valign(0.5)
        grid.set_halign(0.5)
        grid.set_hexpand(True)
        grid.set_vexpand(False)
        row_defs = [
            (
                C_("color sliders panel: red/green/blue: slider label", "R"),
                RGBRedSlider,
                0,
            ), (
                C_("color sliders panel: red/green/blue: slider label", "G"),
                RGBGreenSlider,
                0,
            ), (
                C_("color sliders panel: red/green/blue: slider label", "B"),
                RGBBlueSlider,
                0,
            ), (
                C_("color sliders panel: hue/chroma/luma: slider label", "H"),
                HCYHueSlider,
                12,
            ), (
                C_("color sliders panel: hue/chroma/luma: slider label", "C"),
                HCYChromaSlider,
                0,
            ), (
                C_("color sliders panel: hue/chroma/luma: slider label", "Y"),
                HCYLumaSlider,
                0,
            ),
        ]
        row = 0
        for row_def in row_defs:
            label_text, adj_class, margin_top = row_def
            label = Gtk.Label()
            label.set_text(label_text)
            label.set_tooltip_text(adj_class.STATIC_TOOLTIP_TEXT)
            label.set_vexpand(True)
            label.set_hexpand(False)
            label.set_valign(0.0)
            label.set_margin_top(margin_top)
            label.set_margin_start(3)
            label.set_margin_end(3)
            adj = adj_class()
            adj.set_size_request(100, 22)
            adj.set_vexpand(False)
            adj.set_hexpand(True)
            adj.set_margin_top(margin_top)
            adj.set_margin_left(3)
            adj.set_margin_right(3)
            adj.set_valign(0.0)
            self._sliders.append(adj)
            grid.attach(label, 0, row, 1, 1)
            grid.attach(adj, 1, row, 1, 1)
            row += 1
        align = Gtk.Alignment(
            xalign=0.5, yalign=0.5,
            xscale=1.0, yscale=0.0,
        )
        align.add(grid)
        self._page_widget = align  #: Page's layout widget

    @classmethod
    def get_page_icon_name(self):
        return 'mypaint-tool-component-sliders'

    @classmethod
    def get_page_title(self):
        return C_(
            "color sliders panel: tab title (in tooltip)",
            "Component Sliders"
        )

    @classmethod
    def get_page_description(self):
        return C_(
            "color sliders panel: tab description (in tooltip)",
            "Adjust individual components of the color.",
        )

    def get_page_widget(self):
        return self._page_widget

    def set_color_manager(self, manager):
        ColorAdjuster.set_color_manager(self, manager)
        for slider in self._sliders:
            slider.set_color_manager(manager)

    def render_as_icon(self, cr, size):
        """Renders as an icon into a Cairo context.
        """
        # Strategy: construct tmp R,G,B sliders with a color that shows off
        # their primary a bit. Render carefully (might need special handling
        # for the 16px size).
        from adjbases import ColorManager
        mgr = ColorManager(prefs={}, datapath=".")
        mgr.set_color(RGBColor(0.3, 0.3, 0.4))
        adjs = [RGBRedSlider(), RGBGreenSlider(), RGBBlueSlider()]
        for adj in adjs:
            adj.set_color_manager(mgr)
        if size <= 16:
            cr.save()
            for adj in adjs:
                adj.BORDER_WIDTH = 1
                adj.render_background_cb(cr, wd=16, ht=5)
                cr.translate(0, 5)
            cr.restore()
        else:
            cr.save()
            bar_ht = int(size // 3)
            offset = int((size - bar_ht*3) // 2)
            cr.translate(0, offset)
            for adj in adjs:
                adj.BORDER_WIDTH = max(2, int(size // 16))
                adj.render_background_cb(cr, wd=size, ht=bar_ht)
                cr.translate(0, bar_ht)
            cr.restore()
        for adj in adjs:
            adj.set_color_manager(None)


class RGBRedSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_("color component slider: tooltip", "RGB Red")

    def get_background_validity(self):
        col = self.get_managed_color()
        r, g, b = col.get_rgb()
        return g, b

    def get_color_for_bar_amount(self, amt):
        col = RGBColor(color=self.get_managed_color())
        col.r = amt
        return col

    def get_bar_amount_for_color(self, col):
        return col.r


class RGBGreenSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_("color component slider: tooltip", "RGB Green")

    def get_background_validity(self):
        col = self.get_managed_color()
        r, g, b = col.get_rgb()
        return r, b

    def get_color_for_bar_amount(self, amt):
        col = RGBColor(color=self.get_managed_color())
        col.g = amt
        return col

    def get_bar_amount_for_color(self, col):
        return col.g


class RGBBlueSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_("color component slider: tooltip", "RGB Blue")

    def get_background_validity(self):
        col = self.get_managed_color()
        r, g, b = col.get_rgb()
        return r, g

    def get_color_for_bar_amount(self, amt):
        col = RGBColor(color=self.get_managed_color())
        col.b = amt
        return col

    def get_bar_amount_for_color(self, col):
        return col.b


class HSVHueSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_("color component slider: tooltip", "HSV Hue")
    samples = 4

    def get_color_for_bar_amount(self, amt):
        col = HSVColor(color=self.get_managed_color())
        col.h = amt
        return col

    def get_bar_amount_for_color(self, col):
        return col.h


class HSVSaturationSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_(
        "color component slider: tooltip",
        u"HSV Saturation",
    )

    def get_color_for_bar_amount(self, amt):
        col = HSVColor(color=self.get_managed_color())
        col.s = amt
        return col

    def get_bar_amount_for_color(self, col):
        return col.s


class HSVValueSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_("color component slider: tooltip", "HSV Value")

    def get_color_for_bar_amount(self, amt):
        col = HSVColor(color=self.get_managed_color())
        col.v = amt
        return col

    def get_bar_amount_for_color(self, col):
        return col.v


class HCYHueSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_("color component slider: tooltip", "HCY Hue")
    samples = 4

    def get_color_for_bar_amount(self, amt):
        col = HCYColor(color=self.get_managed_color())
        col.h = amt
        return col

    def get_bar_amount_for_color(self, col):
        col = HCYColor(color=col)
        return col.h


class HCYChromaSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_("color component slider: tooltip", "HCY Chroma")

    def get_color_for_bar_amount(self, amt):
        col = HCYColor(color=self.get_managed_color())
        col.c = amt
        return col

    def get_bar_amount_for_color(self, col):
        col = HCYColor(color=col)
        return col.c


class HCYLumaSlider (SliderColorAdjuster):
    STATIC_TOOLTIP_TEXT = C_(
        "color component slider: tooltip",
        u"HCY Luma (Y')",
    )

    @property
    def samples(self):
        alloc = self.get_allocation()
        len = self.vertical and alloc.height or alloc.width
        len -= self.BORDER_WIDTH * 2
        return min(int(len // 3), 64)

    def get_color_for_bar_amount(self, amt):
        col = HCYColor(color=self.get_managed_color())
        col.y = amt
        return col

    def get_bar_amount_for_color(self, col):
        col = HCYColor(color=col)
        return col.y

    def get_background_validity(self):
        col = HCYColor(color=self.get_managed_color())
        return int(col.h * 1000), int(col.c * 1000)


if __name__ == '__main__':
    import os
    import sys
    from adjbases import ColorManager
    mgr = ColorManager(prefs={}, datapath=".")
    cs_adj = ComponentSlidersAdjusterPage()
    cs_adj.set_color_manager(mgr)
    cs_adj.set_managed_color(RGBColor(0.3, 0.6, 0.7))
    if len(sys.argv) > 1:
        icon_name = cs_adj.get_page_icon_name()
        for dir_name in sys.argv[1:]:
            cs_adj.save_icon_tree(dir_name, icon_name)
    else:
        # Interactive test
        window = Gtk.Window()
        window.add(cs_adj.get_page_widget())
        window.set_title(os.path.basename(sys.argv[0]))
        window.connect("destroy", lambda *a: Gtk.main_quit())
        window.show_all()
        Gtk.main()



# This file is part of MyPaint.
# Copyright (C) 2018 by the Mypaint Development Team
# Copyright (C) 2013 by Andrew Chadwick <a.t.chadwick@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Flood fill tool"""

# Imports
from __future__ import division, print_function

import weakref
from lib.gibindings import Gtk
from lib.gibindings import Pango
from lib.gibindings import GLib
from gettext import gettext as _
from lib.gettext import C_

import cairo

import gui.mode
import gui.cursor
from gui.blendmodehandler import BlendModes
import gui.layers
import gui.overlays

import lib.eotf
import lib.floodfill
import lib.helpers
import lib.mypaintlib
import lib.layer
import lib.modes


# Class defs

class FloodFillMode (
        gui.mode.ScrollableModeMixin, gui.mode.DragMode):
    """Mode for flood-filling with the current brush color"""

    # Class constants

    ACTION_NAME = "FloodFillMode"
    GC_ACTION_NAME = "FloodFillGCMode"

    SPRING_LOADED = False

    permitted_switch_actions = set([
        'RotateViewMode', 'ZoomViewMode', 'PanViewMode',
        'ColorPickMode', 'ShowPopupMenu',
        ])

    _OPTIONS_WIDGET = None
    _CURSOR_FILL_NORMAL = gui.cursor.Name.CROSSHAIR_OPEN_PRECISE
    _CURSOR_FILL_ERASER = gui.cursor.Name.ERASER
    _CURSOR_FILL_ALPHA_LOCKED = gui.cursor.Name.ALPHA_LOCK
    _CURSOR_FILL_COLORIZE = gui.cursor.Name.COLORIZE
    _CURSOR_FILL_FORBIDDEN = gui.cursor.Name.ARROW_FORBIDDEN

    _MODE_CURSORS = [
        _CURSOR_FILL_NORMAL,
        _CURSOR_FILL_ERASER,
        _CURSOR_FILL_ALPHA_LOCKED,
        _CURSOR_FILL_COLORIZE,
    ]

    # Instance vars (and defaults)

    pointer_behavior = gui.mode.Behavior.PAINT_NOBRUSH
    scroll_behavior = gui.mode.Behavior.CHANGE_VIEW

    _current_cursor = (False, _CURSOR_FILL_NORMAL)
    _tdws = None
    _fill_permitted = True
    _x = None
    _y = None

    @property
    def active_cursor(self):
        return self.cursor

    @property
    def inactive_cursor(self):
        return self.cursor

    @property
    def cursor(self):
        gc_on, name = self._current_cursor
        action_name = self.GC_ACTION_NAME if gc_on else self.ACTION_NAME
        return self.app.cursors.get_action_cursor(action_name, name)

    def get_current_cursor(self):
        return self._MODE_CURSORS[self.bm.active_mode.mode_type]

    # Method defs

    def enter(self, doc, **kwds):
        super(FloodFillMode, self).enter(doc, **kwds)
        self._tdws = set([self.doc.tdw])
        self.app.blendmodemanager.register(self.bm)
        rootstack = self.doc.model.layer_stack
        rootstack.current_path_updated += self._update_ui
        rootstack.layer_properties_changed += self._update_ui
        self._update_ui()

    def leave(self, **kwds):
        self.app.blendmodemanager.deregister(self.bm)
        rootstack = self.doc.model.layer_stack
        rootstack.current_path_updated -= self._update_ui
        rootstack.layer_properties_changed -= self._update_ui
        return super(FloodFillMode, self).leave(**kwds)

    @classmethod
    def get_name(cls):
        return C_(
            "Name of the fill mode\n"
            "In other software, 'Flood Fill' is also known as "
            "'Bucket Fill' or just 'Fill'",
            u'Flood Fill'
        )

    def get_usage(self):
        return C_(
            "Usage description of the Flood Fill mode",
            u"Fill areas with color"
        )

    def __init__(self, ignore_modifiers=False, **kwds):
        super(FloodFillMode, self).__init__(**kwds)
        opts = self.get_options_widget()
        self._current_cursor = (opts.gap_closing, self._CURSOR_FILL_NORMAL)
        from gui.application import get_app
        self.app = get_app()
        self.bm = self.get_blend_modes()
        self.bm.mode_changed += self.update_blend_mode
        self._prev_release = (0, None)
        self._seed_pixels = set()
        self._overlay = None
        self._target_pos = None

    def get_blend_modes(self):
        return self.get_options_widget().get_blend_modes()

    def update_blend_mode(self, mode_manager, old_mode, new_mode):
        if old_mode is not new_mode:
            self._update_cursor(self.get_options_widget())

    def drag_update_cb(self, tdw, event, dx, dy):
        """Add pixel coordinate to seed set (if not there already)"""
        x, y = tdw.display_to_model(event.x, event.y)
        seed_candidate = (int(x), int(y))
        self._overlay.add_point((event.x, event.y))
        if self._fill_permitted and seed_candidate not in self._seed_pixels:
            self._seed_pixels.add(seed_candidate)

    def drag_start_cb(self, tdw, event):
        """Create overlay and set initial target pixel & seed"""
        self._seed_pixels = set()
        x, y = tdw.display_to_model(event.x, event.y)
        # Permit setting the target color (the one pixels are tested against)
        # from outside a frame, but no corresponding seed is added
        self._target_pos = (x, y)
        self._overlay = FloodFillOverlay(tdw)
        self._overlay.add_point((event.x, event.y))
        if self._fill_permitted:
            self._seed_pixels.add((int(x), int(y)))

    def drag_stop_cb(self, tdw):
        """Remove overlay and run the fill if valid seeds were marked"""
        self._overlay.cleanup()
        self._overlay = None
        if self._seed_pixels:
            self.fill(tdw)

    def _blend_parameters(self, comp_mode):
        """Get lock_alpha flag and compositing mode"""
        lock_alpha = False

        blend_mode = self.bm.active_mode.mode_type
        # alpha locking - may require two compositing steps per tile
        if blend_mode == BlendModes.LOCK_ALPHA:
            # with CombineNormal, compositing can be done in a single step
            if comp_mode == lib.mypaintlib.CombineNormal:
                comp_mode = lib.mypaintlib.CombineSourceAtop
            lock_alpha = True

        # erasing - overrides other compositing modes when enabled
        elif blend_mode == BlendModes.ERASE:
            comp_mode = lib.mypaintlib.CombineDestinationOut

        # colorize - non-spectral (for now) color blend + alpha locking
        elif blend_mode == BlendModes.COLORIZE:
            comp_mode = lib.mypaintlib.CombineColor
            lock_alpha = True

        return lock_alpha, comp_mode

    def fill(self, tdw):
        """Flood-fill with the current settings and marked seeds

        If the current layer is not fillable, a new layer will always be
        created for the fill.
        """
        self._tdws.add(tdw)
        self._update_ui()
        color = self.doc.app.brush_color_manager.get_color()
        opts = self.get_options_widget()
        make_new_layer = opts.make_new_layer
        rootstack = tdw.doc.layer_stack
        if not rootstack.current.get_fillable():
            make_new_layer = True
        eotf = lib.eotf.eotf()
        rgb = color.get_rgb()
        if eotf != 1.0:
            rgb = (rgb[0]**eotf, rgb[1]**eotf, rgb[2]**eotf)
        view_bbox = None
        if opts.limit_to_view:
            corners = tdw.get_corners_model_coords()
            view_bbox = lib.helpers.rotated_rectangle_bbox(corners)
        seeds = self._seed_pixels
        target_pos = self._target_pos

        lock_alpha, comp_mode = self._blend_parameters(opts.blend_mode)

        fill_args = lib.floodfill.FloodFillArguments(
            target_pos=target_pos,
            seeds=seeds,
            color=rgb,
            tolerance=opts.tolerance,
            offset=opts.offset,
            feather=opts.feather,
            gap_closing_options=opts.gap_closing_options,
            mode=comp_mode,
            lock_alpha=lock_alpha,
            opacity=opts.opacity,
            # Below are set in lib.document
            framed=False,
            bbox=None
        )

        tdw.doc.flood_fill(
            fill_args=fill_args,
            view_bbox=view_bbox,
            sample_merged=opts.sample_merged,
            src_path=opts.src_path,
            make_new_layer=make_new_layer,
            status_cb=status_callback
        )
        opts.make_new_layer = False

    def motion_notify_cb(self, tdw, event):
        """Track position, and update cursor"""
        x, y = tdw.display_to_model(event.x, event.y)
        self._x = x
        self._y = y
        self._tdws.add(tdw)
        self._update_ui()
        return super(FloodFillMode, self).motion_notify_cb(tdw, event)

    def _update_ui(self, *_ignored):
        """Updates the UI from the model"""
        x, y = self._x, self._y
        if None in (x, y):
            x, y = self.current_position()
        model = self.doc.model

        # Determine which layer will receive the fill based on the options
        opts = self.get_options_widget()
        target_layer = model.layer_stack.current
        if opts.make_new_layer:
            target_layer = None

        # Determine whether the target layer can be filled
        permitted = True
        if target_layer is not None:
            permitted = target_layer.visible and not target_layer.locked
        if permitted and model.frame_enabled:
            fx1, fy1, fw, fh = model.get_frame()
            fx2, fy2 = fx1+fw, fy1+fh
            permitted = fx1 <= x < fx2 and fy1 <= y < fy2
        self._fill_permitted = permitted
        self._update_cursor(opts)

    def _update_cursor(self, opts):
        """Update the cursor
        The cursor used is based on fill options, blend mode,
        and if a fill can be run for the current position and layer.
        """
        # Update cursor of any TDWs we've crossed
        if self._fill_permitted:
            cursor = (opts.gap_closing, self.get_current_cursor())
        else:
            cursor = (opts.gap_closing, self._CURSOR_FILL_FORBIDDEN)

        if cursor != self._current_cursor:
            self._current_cursor = cursor
            for tdw in self._tdws:
                tdw.set_override_cursor(self.cursor)

    # Mode options

    def get_options_widget(self):
        """Get the (class singleton) options widget"""
        cls = self.__class__
        if cls._OPTIONS_WIDGET is None:
            widget = FloodFillOptionsWidget()
            cls._OPTIONS_WIDGET = widget
        return cls._OPTIONS_WIDGET


def status_callback(handler):
    """
    Set up and run fill info/cancellation dialog
    :param handler: handler for fill info/cancellation
    :return: False if the fill is cancelled, None otherwise
    """
    app = gui.application.get_app()

    # Prevent escape release (always) triggering mode stack popping
    app.kbm.enabled = False

    # Create new dialog for each occurrence, hopefully
    # occurrences are rare enough for it not to matter very much.
    status_dialog = Gtk.MessageDialog(
        parent=app.drawWindow, buttons=Gtk.ButtonsType.CANCEL)

    curr_stage = [None]

    # Status update ticker callback - also handles dialog destruction
    def status_update():
        if handler.running():
            # Only update the header when the stage has changed
            if curr_stage[0] != handler.stage:
                curr_stage[0] = handler.stage
                status_dialog.set_property("text", handler.stage_string)
            status_dialog.set_property(
                "secondary-text", handler.progress_string)
            return True
        else:
            # Destroy dialog when fill is done, whether cancelled or not
            status_dialog.response(Gtk.ResponseType.OK)
            return False

    # Update the status message 20 times/s
    GLib.timeout_add(50, status_update)
    result = status_dialog.run()
    if result != Gtk.ResponseType.OK:
        handler.cancel()
    status_dialog.hide()
    handler.wait()
    status_dialog.destroy()
    app.kbm.enabled = True
    return result == Gtk.ResponseType.OK


class FloodFillOverlay (gui.overlays.Overlay):
    """
    Overlay indicating pixels that constitute the fill seeds

    The line width is zoom independent, so the visualization
    does not actually correspond 1:1 to the seed pixels.
    Marking pixel locations directly however, would make
    the markings virtually invisible when zoomed out.
    """

    DASH_LENGTH = 9
    # Clean out (SKIP-1)/SKIP points for smoother lines
    SKIP = 6

    def __init__(self, tdw):
        self._tdw = tdw
        self._line_points = []
        self._skip_index = 0
        self._full_bounds = None
        tdw.display_overlays.append(self)

    def cleanup(self):
        self._line_points = []
        self._tdw.display_overlays.remove(self)
        self._skip_index = 0
        self._tdw.queue_draw_area(*self._full_bounds)

    def add_point(self, point):
        x, y = point
        self._full_bounds = self._update_bounds(self._full_bounds, x, y)
        # Clean up lines by only storing every nth point (n = SKIP)
        if len(self._line_points) > 1 and self._skip_index % self.SKIP != 0:
            self._line_points[len(self._line_points) - 1] = point
        else:
            self._line_points.append(point)
        self._skip_index += 1
        self._tdw.queue_draw_area(*self._full_bounds)

    @staticmethod
    def _update_bounds(bounds, x, y):
        if bounds is None:
            return x, y, 1, 1
        else:
            bx, by, w, h = bounds
            if x < bx:
                w += (bx - x)
                bx = x
            elif x > bx:
                w += (x - bx)
            if y < by:
                h += (by - y)
                by = y
            elif y > by:
                h += (y - by)
            return bx, by, w, h

    def paint(self, cr):
        """
        Paint the pixels of the marked set (approximately) as a stroke
        with alternating white and black dashes, for visibility
        """
        d_len = self.DASH_LENGTH
        if len(self._line_points) > 1:
            cr.set_line_join(cairo.LINE_JOIN_ROUND)
            cr.set_source_rgba(0.0, 0.0, 0.0, 0.8)
            cr.set_dash([d_len])
            cr.move_to(*self._line_points[0])
            for (x, y) in self._line_points[1:]:
                cr.line_to(x, y)
            cr.stroke_preserve()
            cr.set_source_rgba(1.0, 1.0, 1.0, 0.8)
            cr.set_dash([d_len], d_len)
            cr.stroke()


class FloodFillOptionsWidget (Gtk.Grid):
    """Configuration widget for the flood fill tool"""

    TOLERANCE_PREF = 'flood_fill.tolerance'
    LIM_TO_VIEW_PREF = 'flood_fill.limit_to_view'
    SAMPLE_MERGED_PREF = 'flood_fill.sample_merged'
    OFFSET_PREF = 'flood_fill.offset'
    FEATHER_PREF = 'flood_fill.feather'
    OPACITY_PREF = 'flood_fill.opacity'
    BLEND_MODE_PREF = 'flood_fill.blend_mode'

    # Gap closing related parameters
    GAP_CLOSING_PREF = 'flood_fill.gap_closing'
    GAP_SIZE_PREF = 'flood_fill.gap_size'
    RETRACT_SEEPS_PREF = 'flood_fill.retract_seeps'
    # "make new layer" is a temporary toggle, and is not saved to prefs

    DEFAULT_TOLERANCE = 0.05
    DEFAULT_LIM_TO_VIEW = False
    DEFAULT_SAMPLE_MERGED = False
    DEFAULT_MAKE_NEW_LAYER = False
    DEFAULT_BLEND_MODE = 0
    DEFAULT_OPACITY = 1.0
    DEFAULT_OFFSET = 0
    DEFAULT_FEATHER = 0

    # Gap closing related defaults
    DEFAULT_GAP_CLOSING = False
    DEFAULT_GAP_SIZE = 5
    DEFAULT_RETRACT_SEEPS = True

    # The state of the blend mode modifiers
    _BLEND_MODES = None

    def update_blend_mode(self, manager, old_mode, new_mode):
        """ Enable/disable mode selection combo box"""
        if old_mode == new_mode:
            return
        if new_mode.mode_type in [BlendModes.ERASE, BlendModes.COLORIZE]:
            self._blend_mode_combo.set_sensitive(False)
        else:
            self._blend_mode_combo.set_sensitive(True)
        self._update_blend_mode_warning(new_mode)

    def _update_blend_mode_warning(self, mode):
        """Show/hide warning label"""
        no_op_combination = (
            mode.mode_type == BlendModes.LOCK_ALPHA and
            self.blend_mode in lib.modes.MODES_DECREASING_BACKDROP_ALPHA
        )
        if no_op_combination and not self._warning_shown:
            self.attach(*self._bm_warning_label)
            self._bm_warning_label[0].show()
            self._warning_shown = True
        elif not no_op_combination and self._warning_shown:
            self.remove(self._bm_warning_label[0])
            self._warning_shown = False
        self.show()

    def __init__(self):
        Gtk.Grid.__init__(self)

        self.set_row_spacing(6)
        self.set_column_spacing(6)
        from gui.application import get_app
        self.app = get_app()
        prefs = self.app.preferences

        row = 0
        label = Gtk.Label()
        label.set_markup(C_(
            "fill options: numeric value that determines whether tested pixels"
            " will be included in the fill, based on color difference",
            u"Tolerance:"))
        label.set_tooltip_text(C_(
            "fill options: Tolerance (tooltip) "
            "Note: 'the start' refers to the color of "
            "the starting point (pixel) of the fill",
            u"How much pixel colors are allowed to vary from the start\n"
            u"before Flood Fill will refuse to fill them"))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self.attach(label, 0, row, 1, 1)
        value = prefs.get(self.TOLERANCE_PREF, self.DEFAULT_TOLERANCE)
        value = float(value)
        adj = Gtk.Adjustment(value=value, lower=0.0, upper=1.0,
                             step_increment=0.05, page_increment=0.05,
                             page_size=0)
        adj.connect("value-changed", self._tolerance_changed_cb)
        self._tolerance_adj = adj
        scale = Gtk.Scale()
        scale.set_hexpand(True)
        scale.set_adjustment(adj)
        scale.set_draw_value(False)
        self.attach(scale, 1, row, 1, 1)

        row += 1
        label = Gtk.Label()
        label.set_markup(C_(
            "fill options: option category (label) "
            "Options under this category relate to what the fill is"
            "based on, not where the actual fill ends up.",
            u"Source:"))
        label.set_tooltip_text(C_(
            "fill options: 'Source:' category (tooltip)",
            u"The input that the fill will be based on"))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self.attach(label, 0, row, 1, 1)

        # Selection independent fill-basis

        root = self.app.doc.model.layer_stack
        src_list = FlatLayerList(root)
        self.src_list = src_list
        combo = Gtk.ComboBox.new_with_model(src_list)
        cell = Gtk.CellRendererText()
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        combo.pack_start(cell, True)

        def layer_name_render(_, name_cell, model, it):
            """
            Display layer groups in italics and child layers
            indented with two spaces per level
            """
            name, path, layer = model[it][:3]
            if name is None:
                name = "Layer"
            if layer is None:
                name_cell.set_property(
                    "markup", "( <i>{text}</i> )".format(text=name)
                )
                return
            indented = "  " * (len(path) - 1) + name
            if isinstance(layer, lib.layer.LayerStack):
                name_cell.set_property(
                    "markup", "<i>{text}</i>".format(text=indented)
                )
            else:
                name_cell.set_property("text", indented)

        def sep_func(model, it):
            return model[it][0] is None

        combo.set_row_separator_func(sep_func)
        combo.set_cell_data_func(cell, layer_name_render)
        combo.set_tooltip_text(
            C_(
                "fill options: 'Source' category: Layer dropdown (tooltip)",
                u"Select a specific layer you want the fill to be based on"
            )
        )
        combo.set_active(0)
        self._prev_src_layer = None
        root.layer_inserted += self._layer_inserted_cb
        self._src_combo_cb_id = combo.connect(
            "changed", self._src_combo_changed_cb
        )
        self._src_combo = combo
        self.attach(combo, 1, row, 2, 1)

        row += 1

        text = C_(
            "fill options: 'Source:' category: toggle (label)\n"
            "When this option is enabled, the fill is based\n"
            "on the combination of all visible layers",
            u"Sample Merged")
        checkbut = Gtk.CheckButton.new_with_label(text)
        checkbut.set_tooltip_text(
            C_("fill options: Sample Merged (tooltip)",
               u"When considering which area to fill, use a\n"
               u"temporary merge of all the visible layers\n"
               u"underneath the current layer")
        )
        self.attach(checkbut, 1, row, 1, 1)
        active = bool(prefs.get(self.SAMPLE_MERGED_PREF,
                                self.DEFAULT_SAMPLE_MERGED))
        checkbut.set_active(active)
        checkbut.connect("toggled", self._sample_merged_toggled_cb)
        self._sample_merged_toggle = checkbut
        self._src_combo.set_sensitive(not active)

        row += 1

        text = C_("fill options: toggle whether the fill will be limited "
                  "by the viewport",
                  u"Limit to View")
        checkbut = Gtk.CheckButton.new_with_label(text)
        checkbut.set_tooltip_text(
            C_("fill options: Limit to View (tooltip)\n"
               "Note: 'that can fit the view' is equivalent to: "
               "'in which the entire visible part of the canvas can fit",
               u"Limit the area that can be filled, based on the viewport.\n"
               u"If the view is rotated, the fill will be limited to the\n"
               u"smallest canvas-aligned rectangle that can fit the view."))
        self.attach(checkbut, 1, row, 1, 1)
        active = bool(prefs.get(self.LIM_TO_VIEW_PREF,
                                self.DEFAULT_LIM_TO_VIEW))
        checkbut.set_active(active)
        checkbut.connect("toggled", self._limit_to_view_toggled_cb)
        self._limit_to_view_toggle = checkbut

        row += 1
        label = Gtk.Label()
        label.set_markup(C_(
            "fill options: option category (label)\n"
            "Options under this category relate to where the fill "
            "will end up (default: the active layer) and how it "
            "will be combined with that layer.",
            u"Target:"))
        label.set_tooltip_text(C_(
            "fill options: 'Target:' category (tooltip)",
            u"Where the output should go"))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self.attach(label, 0, row, 1, 1)

        text = C_(
            "fill options: Target | toggle (label)\n"
            "When this option is enabled, the fill will be placed on a new\n"
            "layer above the active layer. Option resets after each fill.",
            u"New Layer (once)"
        )
        checkbut = Gtk.CheckButton.new_with_label(text)
        checkbut.set_tooltip_text(
            C_("fill options: Target | New Layer (tooltip)",
               u"Create a new layer with the results of the fill.\n"
               u"This is turned off automatically after use."))
        self.attach(checkbut, 1, row, 1, 1)
        active = self.DEFAULT_MAKE_NEW_LAYER
        checkbut.set_active(active)
        self._make_new_layer_toggle = checkbut

        row += 1
        label = Gtk.Label()
        label.set_markup(u"<b>\u26a0</b>")  # unicode warning sign
        label.set_tooltip_text(C_(
            "fill options: Target | Blend Mode dropdown - warning text",
            u"This mode does nothing when alpha locking is enabled!")
        )
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self._warning_shown = False
        # Not attached here, warning label is only shown for no-op combinations
        self._bm_warning_label = (label, 0, row, 1, 1)

        modes = list(lib.modes.STANDARD_MODES)
        modes.remove(lib.mypaintlib.CombineSpectralWGM)
        modes.insert(0, lib.mypaintlib.CombineSpectralWGM)
        combo = gui.layers.new_blend_mode_combo(modes, lib.modes.MODE_STRINGS)
        combo.set_tooltip_text(C_(
            "fill options: Target | Blend Mode dropdown (tooltip)",
            u"Blend mode used when filling"))
        # Reinstate the last _mode id_ independent of mode-list order
        mode_type = prefs.get(self.BLEND_MODE_PREF, self.DEFAULT_BLEND_MODE)
        mode_dict = {mode: index for index, mode, in enumerate(modes)}
        # Fallback is only necessary for compat. if a mode is ever removed
        active = mode_dict.get(int(mode_type), self.DEFAULT_BLEND_MODE)
        combo.set_active(active)
        combo.connect(
            "changed", self._bm_combo_changed_cb
        )
        self._blend_mode_combo = combo
        self.attach(combo, 1, row, 2, 1)

        row += 1
        label = Gtk.Label()
        label.set_markup(_(u"Opacity:"))
        label.set_tooltip_text(C_(
            "fill options: Opacity slider (tooltip)",
            u"Opacity of the fill"))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self.attach(label, 0, row, 1, 1)
        value = prefs.get(self.OPACITY_PREF, self.DEFAULT_OPACITY)
        adj = Gtk.Adjustment(value=value, lower=0.0, upper=1.0,
                             step_increment=0.05, page_increment=0.05,
                             page_size=0)
        adj.connect("value-changed", self._opacity_changed_cb)
        self._opacity_adj = adj
        scale = Gtk.Scale()
        scale.set_hexpand(True)
        scale.set_adjustment(adj)
        scale.set_draw_value(False)
        self.attach(scale, 1, row, 1, 1)

        row += 1
        self.attach(Gtk.Separator(), 0, row, 2, 1)

        row += 1
        label = Gtk.Label()
        label.set_markup(C_(
            "fill options: numeric option - grow/shrink fill (label)",
            u"Offset:"
        ))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self.attach(label, 0, row, 1, 1)

        TILE_SIZE = lib.floodfill.TILE_SIZE
        value = prefs.get(self.OFFSET_PREF, self.DEFAULT_OFFSET)
        adj = Gtk.Adjustment(value=value,
                             lower=-TILE_SIZE, upper=TILE_SIZE,
                             step_increment=1, page_increment=4)
        adj.connect("value-changed", self._offset_changed_cb)
        self._offset_adj = adj
        spinbut = Gtk.SpinButton()
        spinbut.set_tooltip_text(C_(
            "fill options: Offset (tooltip)",
            u"The distance in pixels to grow or shrink the fill"
        ))
        spinbut.set_hexpand(True)
        spinbut.set_adjustment(adj)
        spinbut.set_numeric(True)
        self.attach(spinbut, 1, row, 1, 1)

        row += 1
        label = Gtk.Label()
        label.set_markup(C_(
            "fill options: numeric option for blurring fill (label)",
            u"Feather:"
        ))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self.attach(label, 0, row, 1, 1)

        value = prefs.get(self.FEATHER_PREF, self.DEFAULT_FEATHER)
        adj = Gtk.Adjustment(value=value,
                             lower=0, upper=TILE_SIZE,
                             step_increment=1, page_increment=4)
        adj.connect("value-changed", self._feather_changed_cb)
        self._feather_adj = adj
        spinbut = Gtk.SpinButton()
        spinbut.set_tooltip_text(C_(
            "fill options: Feather (tooltip)",
            u"The amount of blur to apply to the fill"
        ))
        spinbut.set_hexpand(True)
        spinbut.set_adjustment(adj)
        spinbut.set_numeric(True)
        self.attach(spinbut, 1, row, 1, 1)

        row += 1
        self.attach(Gtk.Separator(), 0, row, 2, 1)

        row += 1
        gap_closing_params = Gtk.Grid()
        self._gap_closing_grid = gap_closing_params

        text = C_(
            "fill options: gap detection toggle (label)",
            u'Use Gap Detection'
        )
        checkbut = Gtk.CheckButton.new_with_label(text)
        checkbut.set_tooltip_text(C_(
            "fill options: Use Gap Detection (tooltip)",
            u"Try to detect gaps and not fill past them.\n"
            u"Note: This can be a lot slower than the regular fill, "
            u"only enable when you need it."
        ))
        self._gap_closing_toggle = checkbut
        checkbut.connect("toggled", self._gap_closing_toggled_cb)
        active = prefs.get(self.GAP_CLOSING_PREF, self.DEFAULT_GAP_CLOSING)
        checkbut.set_active(active)
        gap_closing_params.set_sensitive(active)
        self.attach(checkbut, 0, row, 2, 1)

        row += 1
        self.attach(gap_closing_params, 0, row, 2, 1)

        gcp_row = 0
        label = Gtk.Label()
        label.set_markup(C_(
            "fill options: gap-detection sub-option, numeric setting (label)",
            u"Max Gap Size:"
        ))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        gap_closing_params.attach(label, 0, gcp_row, 1, 1)

        value = prefs.get(self.GAP_SIZE_PREF, self.DEFAULT_GAP_SIZE)
        adj = Gtk.Adjustment(value=value,
                             lower=1, upper=int(TILE_SIZE/2),
                             step_increment=1, page_increment=4)
        adj.connect("value-changed", self._max_gap_size_changed_cb)
        self._max_gap_adj = adj
        spinbut = Gtk.SpinButton()
        spinbut.set_tooltip_text(C_(
            "fill options: Max Gap Size (tooltip)",
            u"The size of the largest gaps that can be detected.\n"
            u"Using large values can make the fill run a lot slower."
        ))
        spinbut.set_hexpand(True)
        spinbut.set_adjustment(adj)
        spinbut.set_numeric(True)
        gap_closing_params.attach(spinbut, 1, gcp_row, 1, 1)

        gcp_row += 1
        text = C_(
            "fill options: on/off sub-option, numeric (label)\n"
            "When enabled, if the fill stops after going past a detected gap, "
            "it 'pulls' the fill back out of the gap to the other side of it.",
            u"Prevent seeping"
        )
        checkbut = Gtk.CheckButton.new_with_label(text)
        active = prefs.get(self.RETRACT_SEEPS_PREF, self.DEFAULT_RETRACT_SEEPS)
        checkbut.set_active(active)
        checkbut.set_tooltip_text(C_(
            "fill options: Prevent seeping (tooltip)",
            u"Try to prevent the fill from seeping into the gaps.\n"
            u"If a fill starts in a detected gap, this option will do nothing."
        ))
        checkbut.connect("toggled", self._retract_seeps_toggled_cb)
        self._retract_seeps_toggle = checkbut
        gap_closing_params.attach(checkbut, 1, gcp_row, 1, 1)

        row += 1
        align = Gtk.Alignment.new(0.5, 1.0, 1.0, 0.0)
        align.set_vexpand(True)
        button = Gtk.Button(label=_("Reset"))
        button.connect("clicked", self._reset_clicked_cb)
        button.set_tooltip_text(C_(
            "fill options: Reset button (tooltip)",
            "Reset options to their defaults"))
        align.add(button)
        self.attach(align, 0, row, 2, 1)

        # Set up blend modifier callbacks
        self.bm = self.get_blend_modes()
        self.bm.mode_changed += self.update_blend_mode

    # Fill blend modes
    def get_blend_modes(self):
        """Get the (class singleton) blend modes manager"""
        cls = self.__class__
        if cls._BLEND_MODES is None:
            cls._BLEND_MODES = BlendModes()
        return cls._BLEND_MODES

    @property
    def tolerance(self):
        return float(self._tolerance_adj.get_value())

    @property
    def opacity(self):
        return float(self._opacity_adj.get_value())

    @property
    def make_new_layer(self):
        return bool(self._make_new_layer_toggle.get_active())

    @make_new_layer.setter
    def make_new_layer(self, value):
        self._make_new_layer_toggle.set_active(bool(value))

    @property
    def blend_mode(self):
        active = self._blend_mode_combo.get_active()
        active = 0 if active == -1 else active
        return self._blend_mode_combo.get_model()[active][0]

    @property
    def sample_merged(self):
        return bool(self._sample_merged_toggle.get_active())

    @property
    def limit_to_view(self):
        return bool(self._limit_to_view_toggle.get_active())

    @property
    def src_path(self):
        row = self._src_combo.get_active_iter()
        if row is not None:
            return self._src_combo.get_model()[row][1]
        else:
            return None

    @property
    def offset(self):
        return int(self._offset_adj.get_value())

    @property
    def feather(self):
        return int(self._feather_adj.get_value())

    @property
    def gap_closing(self):
        return bool(self._gap_closing_toggle.get_active())

    @property
    def max_gap_size(self):
        return int(self._max_gap_adj.get_value())

    @property
    def retract_seeps(self):
        return bool(self._retract_seeps_toggle.get_active())

    @property
    def gap_closing_options(self):
        if self.gap_closing:
            return lib.floodfill.GapClosingOptions(
                self.max_gap_size, self.retract_seeps)
        else:
            return None

    def _tolerance_changed_cb(self, adj):
        self.app.preferences[self.TOLERANCE_PREF] = self.tolerance

    def _limit_to_view_toggled_cb(self, checkbut):
        self.app.preferences[self.LIM_TO_VIEW_PREF] = self.limit_to_view

    def _sample_merged_toggled_cb(self, checkbut):
        self._src_combo.set_sensitive(not self.sample_merged)
        self.app.preferences[self.SAMPLE_MERGED_PREF] = self.sample_merged

    def _opacity_changed_cb(self, adj):
        self.app.preferences[self.OPACITY_PREF] = self.opacity

    def _offset_changed_cb(self, adj):
        self.app.preferences[self.OFFSET_PREF] = self.offset

    def _feather_changed_cb(self, adj):
        self.app.preferences[self.FEATHER_PREF] = self.feather

    def _gap_closing_toggled_cb(self, adj):
        self._gap_closing_grid.set_sensitive(self.gap_closing)
        self.app.preferences[self.GAP_CLOSING_PREF] = self.gap_closing

    def _max_gap_size_changed_cb(self, adj):
        self.app.preferences[self.GAP_SIZE_PREF] = self.max_gap_size

    def _retract_seeps_toggled_cb(self, adj):
        self.app.preferences[self.RETRACT_SEEPS_PREF] = self.retract_seeps

    def _bm_combo_changed_cb(self, combo):
        self.app.preferences[self.BLEND_MODE_PREF] = self.blend_mode
        self._update_blend_mode_warning(self.get_blend_modes().active_mode)

    def _layer_inserted_cb(self, root, path):
        """Check if the newly inserted layer was the last actively
        selected fill src layer, and reinstate the selection if so.
        """
        layer = root.deepget(path)
        if layer and self._prev_src_layer and self._prev_src_layer() is layer:
            # Restore previous selection layer
            combo = self._src_combo
            for entry in combo.get_model():
                if entry[2] is layer:
                    # Don't trigger callback
                    with combo.handler_block(self._src_combo_cb_id):
                        combo.set_active_iter(entry.iter)
                    return

    def _src_combo_changed_cb(self, combo):
        """Track the last selected choice of layer to maintain
        selection between layer moves that use intermediate
        layer deletion (as well undoing layer deletions)
        """
        row = combo.get_active_iter()
        if row is not None:
            layer = combo.get_model()[row][2]
            if layer is None:
                self._prev_src_layer = None
            else:
                self._prev_src_layer = weakref.ref(layer)
        else:
            # Option unset by layer deletion, set to default
            # without triggering callback again
            with combo.handler_block(self._src_combo_cb_id):
                combo.set_active(0)

    def _reset_clicked_cb(self, button):
        self._tolerance_adj.set_value(self.DEFAULT_TOLERANCE)
        self._make_new_layer_toggle.set_active(self.DEFAULT_MAKE_NEW_LAYER)
        self._src_combo.set_active(0)
        self._limit_to_view_toggle.set_active(self.DEFAULT_LIM_TO_VIEW)
        self._sample_merged_toggle.set_active(self.DEFAULT_SAMPLE_MERGED)
        self._opacity_adj.set_value(self.DEFAULT_OPACITY)
        self._offset_adj.set_value(self.DEFAULT_OFFSET)
        self._feather_adj.set_value(self.DEFAULT_FEATHER)
        # Gap closing params
        self._max_gap_adj.set_value(self.DEFAULT_GAP_SIZE)
        self._retract_seeps_toggle.set_active(self.DEFAULT_RETRACT_SEEPS)
        self._gap_closing_toggle.set_active(self.DEFAULT_GAP_CLOSING)


class FlatLayerList(Gtk.ListStore):
    """Stores a flattened copy of the layer tree"""

    def __init__(self, root_stack):
        super(FlatLayerList, self).__init__()

        root_stack.layer_properties_changed += self._layer_props_changed_cb
        root_stack.layer_inserted += self._layer_inserted_cb
        root_stack.layer_deleted += self._layer_deleted_cb

        self.root = root_stack
        # Column data : name, layer_path, layer
        self.set_column_types((str, object, object))
        default_selection = C_(
            "fill option: default option in the Source Layer dropdown",
            u"Selected Layer"
        )
        # Add default option and separator
        self.append((default_selection, None, None))
        self.append((None, None, None))
        # Flatten layer tree into rows
        for layer in root_stack:
            self._initalize(layer)

    def _layer_props_changed_cb(self, root, layerpath, layer, changed):
        """Update copies of layer names when changed"""
        if 'name' in changed:
            for item in self:
                if item[1] == layerpath:
                    item[0] = layer.name
                    return

    def _layer_inserted_cb(self, root, path):
        """Create a row for the inserted layer and update
        the paths of existing layers if necessary
        """
        layer = root.deepget(path)
        new_row = (layer.name, path, layer)
        for item in self:
            item_path = item[1]
            if item_path and path <= item_path:
                row_iter = item.iter
                self.insert_before(row_iter, new_row)
                self._update_paths(row_iter)
                return
        # If layer added to bottom, no other updates necessary
        self.append(new_row)

    def _layer_deleted_cb(self, root, path):
        """Remove the row for the deleted layer, and also any
        rows for layers that were children of the deleted layer
        """
        def is_child(p):
            return lib.layer.path_startswith(p, path)

        for item in self:
            if item[1] == path:
                row = item.iter
                # Remove rows for all children
                while self.remove(row) and is_child(self[row][1]):
                    pass
                # Update rows (if any) below last deleted
                if self.iter_is_valid(row):
                    self._update_paths(row)
                return

    def _update_paths(self, row_iter):
        """Update the paths for existing layers
        at or below the point of the given iterator
        """
        while row_iter:
            item = self[row_iter]
            path = self.root.deepindex(item[2])
            if path is not None:
                item[1] = path
            row_iter = self.iter_next(row_iter)

    def _initalize(self, layer):
        """Add a new row for the layer (unless it is the root)
        Subtrees are added recursively, depth-first traversal.
        """
        # if layer is not self.root:
        name = layer.name
        path = self.root.deepindex(layer)
        self.append((name, path, layer))
        if isinstance(layer, lib.layer.LayerStack):
            for child in layer:
                self._initalize(child)



# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.4.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if getattr(globals().get("__spec__"), "parent", None) or __package__ or "." in __name__:
    from . import _mypaintlib
else:
    import _mypaintlib

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


class SwigPyIterator(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _mypaintlib.delete_SwigPyIterator

    def value(self):
        return _mypaintlib.SwigPyIterator_value(self)

    def incr(self, n=1):
        return _mypaintlib.SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        return _mypaintlib.SwigPyIterator_decr(self, n)

    def distance(self, x):
        return _mypaintlib.SwigPyIterator_distance(self, x)

    def equal(self, x):
        return _mypaintlib.SwigPyIterator_equal(self, x)

    def copy(self):
        return _mypaintlib.SwigPyIterator_copy(self)

    def next(self):
        return _mypaintlib.SwigPyIterator_next(self)

    def __next__(self):
        return _mypaintlib.SwigPyIterator___next__(self)

    def previous(self):
        return _mypaintlib.SwigPyIterator_previous(self)

    def advance(self, n):
        return _mypaintlib.SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        return _mypaintlib.SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        return _mypaintlib.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        return _mypaintlib.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        return _mypaintlib.SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        return _mypaintlib.SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        return _mypaintlib.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self

# Register SwigPyIterator in _mypaintlib:
_mypaintlib.SwigPyIterator_swigregister(SwigPyIterator)
cvar = _mypaintlib.cvar
heavy_debug = cvar.heavy_debug

class IntVector(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self):
        return _mypaintlib.IntVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _mypaintlib.IntVector___nonzero__(self)

    def __bool__(self):
        return _mypaintlib.IntVector___bool__(self)

    def __len__(self):
        return _mypaintlib.IntVector___len__(self)

    def __getslice__(self, i, j):
        return _mypaintlib.IntVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _mypaintlib.IntVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _mypaintlib.IntVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _mypaintlib.IntVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _mypaintlib.IntVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _mypaintlib.IntVector___setitem__(self, *args)

    def pop(self):
        return _mypaintlib.IntVector_pop(self)

    def append(self, x):
        return _mypaintlib.IntVector_append(self, x)

    def empty(self):
        return _mypaintlib.IntVector_empty(self)

    def size(self):
        return _mypaintlib.IntVector_size(self)

    def swap(self, v):
        return _mypaintlib.IntVector_swap(self, v)

    def begin(self):
        return _mypaintlib.IntVector_begin(self)

    def end(self):
        return _mypaintlib.IntVector_end(self)

    def rbegin(self):
        return _mypaintlib.IntVector_rbegin(self)

    def rend(self):
        return _mypaintlib.IntVector_rend(self)

    def clear(self):
        return _mypaintlib.IntVector_clear(self)

    def get_allocator(self):
        return _mypaintlib.IntVector_get_allocator(self)

    def pop_back(self):
        return _mypaintlib.IntVector_pop_back(self)

    def erase(self, *args):
        return _mypaintlib.IntVector_erase(self, *args)

    def __init__(self, *args):
        _mypaintlib.IntVector_swiginit(self, _mypaintlib.new_IntVector(*args))

    def push_back(self, x):
        return _mypaintlib.IntVector_push_back(self, x)

    def front(self):
        return _mypaintlib.IntVector_front(self)

    def back(self):
        return _mypaintlib.IntVector_back(self)

    def assign(self, n, x):
        return _mypaintlib.IntVector_assign(self, n, x)

    def resize(self, *args):
        return _mypaintlib.IntVector_resize(self, *args)

    def insert(self, *args):
        return _mypaintlib.IntVector_insert(self, *args)

    def reserve(self, n):
        return _mypaintlib.IntVector_reserve(self, n)

    def capacity(self):
        return _mypaintlib.IntVector_capacity(self)
    __swig_destroy__ = _mypaintlib.delete_IntVector

# Register IntVector in _mypaintlib:
_mypaintlib.IntVector_swigregister(IntVector)
class RectVector(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self):
        return _mypaintlib.RectVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _mypaintlib.RectVector___nonzero__(self)

    def __bool__(self):
        return _mypaintlib.RectVector___bool__(self)

    def __len__(self):
        return _mypaintlib.RectVector___len__(self)

    def __getslice__(self, i, j):
        return _mypaintlib.RectVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _mypaintlib.RectVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _mypaintlib.RectVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _mypaintlib.RectVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _mypaintlib.RectVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _mypaintlib.RectVector___setitem__(self, *args)

    def pop(self):
        return _mypaintlib.RectVector_pop(self)

    def append(self, x):
        return _mypaintlib.RectVector_append(self, x)

    def empty(self):
        return _mypaintlib.RectVector_empty(self)

    def size(self):
        return _mypaintlib.RectVector_size(self)

    def swap(self, v):
        return _mypaintlib.RectVector_swap(self, v)

    def begin(self):
        return _mypaintlib.RectVector_begin(self)

    def end(self):
        return _mypaintlib.RectVector_end(self)

    def rbegin(self):
        return _mypaintlib.RectVector_rbegin(self)

    def rend(self):
        return _mypaintlib.RectVector_rend(self)

    def clear(self):
        return _mypaintlib.RectVector_clear(self)

    def get_allocator(self):
        return _mypaintlib.RectVector_get_allocator(self)

    def pop_back(self):
        return _mypaintlib.RectVector_pop_back(self)

    def erase(self, *args):
        return _mypaintlib.RectVector_erase(self, *args)

    def __init__(self, *args):
        _mypaintlib.RectVector_swiginit(self, _mypaintlib.new_RectVector(*args))

    def push_back(self, x):
        return _mypaintlib.RectVector_push_back(self, x)

    def front(self):
        return _mypaintlib.RectVector_front(self)

    def back(self):
        return _mypaintlib.RectVector_back(self)

    def assign(self, n, x):
        return _mypaintlib.RectVector_assign(self, n, x)

    def resize(self, *args):
        return _mypaintlib.RectVector_resize(self, *args)

    def insert(self, *args):
        return _mypaintlib.RectVector_insert(self, *args)

    def reserve(self, n):
        return _mypaintlib.RectVector_reserve(self, n)

    def capacity(self):
        return _mypaintlib.RectVector_capacity(self)
    __swig_destroy__ = _mypaintlib.delete_RectVector

# Register RectVector in _mypaintlib:
_mypaintlib.RectVector_swigregister(RectVector)
class DoubleVector(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self):
        return _mypaintlib.DoubleVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _mypaintlib.DoubleVector___nonzero__(self)

    def __bool__(self):
        return _mypaintlib.DoubleVector___bool__(self)

    def __len__(self):
        return _mypaintlib.DoubleVector___len__(self)

    def __getslice__(self, i, j):
        return _mypaintlib.DoubleVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _mypaintlib.DoubleVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _mypaintlib.DoubleVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _mypaintlib.DoubleVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _mypaintlib.DoubleVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _mypaintlib.DoubleVector___setitem__(self, *args)

    def pop(self):
        return _mypaintlib.DoubleVector_pop(self)

    def append(self, x):
        return _mypaintlib.DoubleVector_append(self, x)

    def empty(self):
        return _mypaintlib.DoubleVector_empty(self)

    def size(self):
        return _mypaintlib.DoubleVector_size(self)

    def swap(self, v):
        return _mypaintlib.DoubleVector_swap(self, v)

    def begin(self):
        return _mypaintlib.DoubleVector_begin(self)

    def end(self):
        return _mypaintlib.DoubleVector_end(self)

    def rbegin(self):
        return _mypaintlib.DoubleVector_rbegin(self)

    def rend(self):
        return _mypaintlib.DoubleVector_rend(self)

    def clear(self):
        return _mypaintlib.DoubleVector_clear(self)

    def get_allocator(self):
        return _mypaintlib.DoubleVector_get_allocator(self)

    def pop_back(self):
        return _mypaintlib.DoubleVector_pop_back(self)

    def erase(self, *args):
        return _mypaintlib.DoubleVector_erase(self, *args)

    def __init__(self, *args):
        _mypaintlib.DoubleVector_swiginit(self, _mypaintlib.new_DoubleVector(*args))

    def push_back(self, x):
        return _mypaintlib.DoubleVector_push_back(self, x)

    def front(self):
        return _mypaintlib.DoubleVector_front(self)

    def back(self):
        return _mypaintlib.DoubleVector_back(self)

    def assign(self, n, x):
        return _mypaintlib.DoubleVector_assign(self, n, x)

    def resize(self, *args):
        return _mypaintlib.DoubleVector_resize(self, *args)

    def insert(self, *args):
        return _mypaintlib.DoubleVector_insert(self, *args)

    def reserve(self, n):
        return _mypaintlib.DoubleVector_reserve(self, n)

    def capacity(self):
        return _mypaintlib.DoubleVector_capacity(self)
    __swig_destroy__ = _mypaintlib.delete_DoubleVector

# Register DoubleVector in _mypaintlib:
_mypaintlib.DoubleVector_swigregister(DoubleVector)
class Rect(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    x = property(_mypaintlib.Rect_x_get, _mypaintlib.Rect_x_set)
    y = property(_mypaintlib.Rect_y_get, _mypaintlib.Rect_y_set)
    w = property(_mypaintlib.Rect_w_get, _mypaintlib.Rect_w_set)
    h = property(_mypaintlib.Rect_h_get, _mypaintlib.Rect_h_set)

    def __init__(self):
        _mypaintlib.Rect_swiginit(self, _mypaintlib.new_Rect())
    __swig_destroy__ = _mypaintlib.delete_Rect

# Register Rect in _mypaintlib:
_mypaintlib.Rect_swigregister(Rect)
class Surface(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _mypaintlib.delete_Surface

    def get_surface_interface(self):
        return _mypaintlib.Surface_get_surface_interface(self)

    def get_surface2_interface(self):
        return _mypaintlib.Surface_get_surface2_interface(self)

# Register Surface in _mypaintlib:
_mypaintlib.Surface_swigregister(Surface)
class Brush(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self):
        _mypaintlib.Brush_swiginit(self, _mypaintlib.new_Brush())
    __swig_destroy__ = _mypaintlib.delete_Brush

    def reset(self):
        return _mypaintlib.Brush_reset(self)

    def new_stroke(self):
        return _mypaintlib.Brush_new_stroke(self)

    def set_base_value(self, id, value):
        return _mypaintlib.Brush_set_base_value(self, id, value)

    def set_mapping_n(self, id, input, n):
        return _mypaintlib.Brush_set_mapping_n(self, id, input, n)

    def set_mapping_point(self, id, input, index, x, y):
        return _mypaintlib.Brush_set_mapping_point(self, id, input, index, x, y)

    def get_state(self, i):
        return _mypaintlib.Brush_get_state(self, i)

    def set_state(self, i, value):
        return _mypaintlib.Brush_set_state(self, i, value)

    def stroke_to(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation):
        return _mypaintlib.Brush_stroke_to(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation)

    def stroke_to_linear(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation):
        return _mypaintlib.Brush_stroke_to_linear(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation)

    def get_total_stroke_painting_time(self):
        return _mypaintlib.Brush_get_total_stroke_painting_time(self)

    def set_print_inputs(self, enabled):
        return _mypaintlib.Brush_set_print_inputs(self, enabled)

# Register Brush in _mypaintlib:
_mypaintlib.Brush_swigregister(Brush)
class MappingWrapper(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, inputs_):
        _mypaintlib.MappingWrapper_swiginit(self, _mypaintlib.new_MappingWrapper(inputs_))
    __swig_destroy__ = _mypaintlib.delete_MappingWrapper

    def set_n(self, input, n):
        return _mypaintlib.MappingWrapper_set_n(self, input, n)

    def set_point(self, input, index, x, y):
        return _mypaintlib.MappingWrapper_set_point(self, input, index, x, y)

    def is_constant(self):
        return _mypaintlib.MappingWrapper_is_constant(self)

    def calculate(self, data):
        return _mypaintlib.MappingWrapper_calculate(self, data)

    def calculate_single_input(self, input):
        return _mypaintlib.MappingWrapper_calculate_single_input(self, input)

# Register MappingWrapper in _mypaintlib:
_mypaintlib.MappingWrapper_swigregister(MappingWrapper)
class PythonBrush(Brush):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def get_states_as_array(self):
        return _mypaintlib.PythonBrush_get_states_as_array(self)

    def set_states_from_array(self, obj):
        return _mypaintlib.PythonBrush_set_states_from_array(self, obj)

    def stroke_to(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation):
        return _mypaintlib.PythonBrush_stroke_to(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation)

    def stroke_to_linear(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation):
        return _mypaintlib.PythonBrush_stroke_to_linear(self, surface, x, y, pressure, xtilt, ytilt, dtime, viewzoom, viewrotation, barrel_rotation)

    def __init__(self):
        _mypaintlib.PythonBrush_swiginit(self, _mypaintlib.new_PythonBrush())
    __swig_destroy__ = _mypaintlib.delete_PythonBrush

# Register PythonBrush in _mypaintlib:
_mypaintlib.PythonBrush_swigregister(PythonBrush)
BBOXES = _mypaintlib.BBOXES
SymmetryVertical = _mypaintlib.SymmetryVertical
SymmetryHorizontal = _mypaintlib.SymmetryHorizontal
SymmetryVertHorz = _mypaintlib.SymmetryVertHorz
SymmetryRotational = _mypaintlib.SymmetryRotational
SymmetrySnowflake = _mypaintlib.SymmetrySnowflake
NumSymmetryTypes = _mypaintlib.NumSymmetryTypes
class TiledSurface(Surface):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, self_):
        _mypaintlib.TiledSurface_swiginit(self, _mypaintlib.new_TiledSurface(self_))
    __swig_destroy__ = _mypaintlib.delete_TiledSurface

    def set_symmetry_state(self, active, center_x, center_y, symmetry_type, rot_symmetry_lines):
        return _mypaintlib.TiledSurface_set_symmetry_state(self, active, center_x, center_y, symmetry_type, rot_symmetry_lines)

    def begin_atomic(self):
        return _mypaintlib.TiledSurface_begin_atomic(self)

    def end_atomic(self):
        return _mypaintlib.TiledSurface_end_atomic(self)

    def draw_dab(self, x, y, radius, color_r, color_g, color_b, opaque, hardness=0.5, color_a=1.0, aspect_ratio=1.0, angle=0.0, lock_alpha=0.0, colorize=0.0, posterize=0.0, posterize_num=0.0, paint=1.0):
        return _mypaintlib.TiledSurface_draw_dab(self, x, y, radius, color_r, color_g, color_b, opaque, hardness, color_a, aspect_ratio, angle, lock_alpha, colorize, posterize, posterize_num, paint)

    def get_color(self, x, y, radius):
        return _mypaintlib.TiledSurface_get_color(self, x, y, radius)

    def get_alpha(self, x, y, radius):
        return _mypaintlib.TiledSurface_get_alpha(self, x, y, radius)

    def get_surface_interface(self):
        return _mypaintlib.TiledSurface_get_surface_interface(self)

    def get_surface2_interface(self):
        return _mypaintlib.TiledSurface_get_surface2_interface(self)

# Register TiledSurface in _mypaintlib:
_mypaintlib.TiledSurface_swigregister(TiledSurface)
TILE_SIZE = cvar.TILE_SIZE
MAX_MIPMAP_LEVEL = cvar.MAX_MIPMAP_LEVEL


def get_module(name):
    return _mypaintlib.get_module(name)

def new_py_tiled_surface(pModule):
    return _mypaintlib.new_py_tiled_surface(pModule)

def mypaint_python_surface_factory(user_data):
    return _mypaintlib.mypaint_python_surface_factory(user_data)

def tile_downscale_rgba16(src, dst, dst_x, dst_y):
    return _mypaintlib.tile_downscale_rgba16(src, dst, dst_x, dst_y)

def tile_copy_rgba16_into_rgba16(src, dst):
    return _mypaintlib.tile_copy_rgba16_into_rgba16(src, dst)

def tile_clear_rgba16(dst):
    return _mypaintlib.tile_clear_rgba16(dst)

def tile_clear_rgba8(dst):
    return _mypaintlib.tile_clear_rgba8(dst)

def tile_convert_rgba16_to_rgba8(src, dst, EOTF):
    return _mypaintlib.tile_convert_rgba16_to_rgba8(src, dst, EOTF)

def tile_convert_rgbu16_to_rgbu8(src, dst, EOTF):
    return _mypaintlib.tile_convert_rgbu16_to_rgbu8(src, dst, EOTF)

def tile_convert_rgba8_to_rgba16(src, dst, EOTF):
    return _mypaintlib.tile_convert_rgba8_to_rgba16(src, dst, EOTF)

def tile_rgba2flat(dst_obj, bg_obj):
    return _mypaintlib.tile_rgba2flat(dst_obj, bg_obj)

def tile_flat2rgba(dst_obj, bg_obj):
    return _mypaintlib.tile_flat2rgba(dst_obj, bg_obj)

def tile_perceptual_change_strokemap(a_obj, b_obj, res_obj):
    return _mypaintlib.tile_perceptual_change_strokemap(a_obj, b_obj, res_obj)
CombineNormal = _mypaintlib.CombineNormal
CombineMultiply = _mypaintlib.CombineMultiply
CombineScreen = _mypaintlib.CombineScreen
CombineOverlay = _mypaintlib.CombineOverlay
CombineDarken = _mypaintlib.CombineDarken
CombineLighten = _mypaintlib.CombineLighten
CombineHardLight = _mypaintlib.CombineHardLight
CombineSoftLight = _mypaintlib.CombineSoftLight
CombineColorBurn = _mypaintlib.CombineColorBurn
CombineColorDodge = _mypaintlib.CombineColorDodge
CombineDifference = _mypaintlib.CombineDifference
CombineExclusion = _mypaintlib.CombineExclusion
CombineHue = _mypaintlib.CombineHue
CombineSaturation = _mypaintlib.CombineSaturation
CombineColor = _mypaintlib.CombineColor
CombineLuminosity = _mypaintlib.CombineLuminosity
CombineLighter = _mypaintlib.CombineLighter
CombineDestinationIn = _mypaintlib.CombineDestinationIn
CombineDestinationOut = _mypaintlib.CombineDestinationOut
CombineSourceAtop = _mypaintlib.CombineSourceAtop
CombineDestinationAtop = _mypaintlib.CombineDestinationAtop
CombineSpectralWGM = _mypaintlib.CombineSpectralWGM
NumCombineModes = _mypaintlib.NumCombineModes

def combine_mode_get_info(mode):
    return _mypaintlib.combine_mode_get_info(mode)

def tile_combine(mode, src_obj, dst_obj, dst_has_alpha, src_opacity):
    return _mypaintlib.tile_combine(mode, src_obj, dst_obj, dst_has_alpha, src_opacity)
class SCWSColorSelector(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def get_size(self):
        return _mypaintlib.SCWSColorSelector_get_size(self)
    brush_h = property(_mypaintlib.SCWSColorSelector_brush_h_get, _mypaintlib.SCWSColorSelector_brush_h_set)
    brush_s = property(_mypaintlib.SCWSColorSelector_brush_s_get, _mypaintlib.SCWSColorSelector_brush_s_set)
    brush_v = property(_mypaintlib.SCWSColorSelector_brush_v_get, _mypaintlib.SCWSColorSelector_brush_v_set)

    def set_brush_color(self, h, s, v):
        return _mypaintlib.SCWSColorSelector_set_brush_color(self, h, s, v)

    def get_hsva_at(self, h, s, v, a, x, y, adjust_color=True, only_colors=True, mark_h=0.0):
        return _mypaintlib.SCWSColorSelector_get_hsva_at(self, h, s, v, a, x, y, adjust_color, only_colors, mark_h)

    def pick_color_at(self, x, y):
        return _mypaintlib.SCWSColorSelector_pick_color_at(self, x, y)

    def render(self, obj):
        return _mypaintlib.SCWSColorSelector_render(self, obj)

    def __init__(self):
        _mypaintlib.SCWSColorSelector_swiginit(self, _mypaintlib.new_SCWSColorSelector())
    __swig_destroy__ = _mypaintlib.delete_SCWSColorSelector

# Register SCWSColorSelector in _mypaintlib:
_mypaintlib.SCWSColorSelector_swigregister(SCWSColorSelector)
colorring_size = cvar.colorring_size
center = cvar.center
RAD_TO_ONE = cvar.RAD_TO_ONE
TWO_PI = cvar.TWO_PI
ONE_OVER_THREE = cvar.ONE_OVER_THREE
TWO_OVER_THREE = cvar.TWO_OVER_THREE

class ColorChangerWash(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    brush_h = property(_mypaintlib.ColorChangerWash_brush_h_get, _mypaintlib.ColorChangerWash_brush_h_set)
    brush_s = property(_mypaintlib.ColorChangerWash_brush_s_get, _mypaintlib.ColorChangerWash_brush_s_set)
    brush_v = property(_mypaintlib.ColorChangerWash_brush_v_get, _mypaintlib.ColorChangerWash_brush_v_set)

    def set_brush_color(self, h, s, v):
        return _mypaintlib.ColorChangerWash_set_brush_color(self, h, s, v)

    def get_size(self):
        return _mypaintlib.ColorChangerWash_get_size(self)

    def render(self, obj):
        return _mypaintlib.ColorChangerWash_render(self, obj)

    def pick_color_at(self, x_, y_):
        return _mypaintlib.ColorChangerWash_pick_color_at(self, x_, y_)

    def __init__(self):
        _mypaintlib.ColorChangerWash_swiginit(self, _mypaintlib.new_ColorChangerWash())
    __swig_destroy__ = _mypaintlib.delete_ColorChangerWash

# Register ColorChangerWash in _mypaintlib:
_mypaintlib.ColorChangerWash_swigregister(ColorChangerWash)
ccw_size = cvar.ccw_size
v06_colorchanger = cvar.v06_colorchanger

class ColorChangerCrossedBowl(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    brush_h = property(_mypaintlib.ColorChangerCrossedBowl_brush_h_get, _mypaintlib.ColorChangerCrossedBowl_brush_h_set)
    brush_s = property(_mypaintlib.ColorChangerCrossedBowl_brush_s_get, _mypaintlib.ColorChangerCrossedBowl_brush_s_set)
    brush_v = property(_mypaintlib.ColorChangerCrossedBowl_brush_v_get, _mypaintlib.ColorChangerCrossedBowl_brush_v_set)

    def set_brush_color(self, h, s, v):
        return _mypaintlib.ColorChangerCrossedBowl_set_brush_color(self, h, s, v)

    def get_size(self):
        return _mypaintlib.ColorChangerCrossedBowl_get_size(self)

    def render(self, obj):
        return _mypaintlib.ColorChangerCrossedBowl_render(self, obj)

    def pick_color_at(self, x_, y_):
        return _mypaintlib.ColorChangerCrossedBowl_pick_color_at(self, x_, y_)

    def __init__(self):
        _mypaintlib.ColorChangerCrossedBowl_swiginit(self, _mypaintlib.new_ColorChangerCrossedBowl())
    __swig_destroy__ = _mypaintlib.delete_ColorChangerCrossedBowl

# Register ColorChangerCrossedBowl in _mypaintlib:
_mypaintlib.ColorChangerCrossedBowl_swigregister(ColorChangerCrossedBowl)
ccdb_size = cvar.ccdb_size

class ProgressivePNGWriter(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, file, w, h, has_alpha, save_srgb_chunks):
        _mypaintlib.ProgressivePNGWriter_swiginit(self, _mypaintlib.new_ProgressivePNGWriter(file, w, h, has_alpha, save_srgb_chunks))

    def write(self, arr):
        return _mypaintlib.ProgressivePNGWriter_write(self, arr)

    def close(self):
        return _mypaintlib.ProgressivePNGWriter_close(self)
    __swig_destroy__ = _mypaintlib.delete_ProgressivePNGWriter

# Register ProgressivePNGWriter in _mypaintlib:
_mypaintlib.ProgressivePNGWriter_swigregister(ProgressivePNGWriter)

def load_png_fast_progressive(filename, get_buffer_callback, convert_to_srgb):
    return _mypaintlib.load_png_fast_progressive(filename, get_buffer_callback, convert_to_srgb)
class ConstTiles(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    @staticmethod
    def ALPHA_OPAQUE():
        return _mypaintlib.ConstTiles_ALPHA_OPAQUE()

    @staticmethod
    def ALPHA_TRANSPARENT():
        return _mypaintlib.ConstTiles_ALPHA_TRANSPARENT()

    def __init__(self):
        _mypaintlib.ConstTiles_swiginit(self, _mypaintlib.new_ConstTiles())
    __swig_destroy__ = _mypaintlib.delete_ConstTiles

# Register ConstTiles in _mypaintlib:
_mypaintlib.ConstTiles_swigregister(ConstTiles)
class edges(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    north = _mypaintlib.edges_north
    east = _mypaintlib.edges_east
    south = _mypaintlib.edges_south
    west = _mypaintlib.edges_west
    none = _mypaintlib.edges_none

    def __init__(self):
        _mypaintlib.edges_swiginit(self, _mypaintlib.new_edges())
    __swig_destroy__ = _mypaintlib.delete_edges

# Register edges in _mypaintlib:
_mypaintlib.edges_swigregister(edges)
class Filler(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, targ_r, targ_g, targ_b, targ_a, tol):
        _mypaintlib.Filler_swiginit(self, _mypaintlib.new_Filler(targ_r, targ_g, targ_b, targ_a, tol))

    def fill(self, src, dst, seeds, direction, min_x, min_y, max_x, max_y):
        return _mypaintlib.Filler_fill(self, src, dst, seeds, direction, min_x, min_y, max_x, max_y)

    def flood(self, src, dst):
        return _mypaintlib.Filler_flood(self, src, dst)

    def tile_uniformity(self, is_empty, src):
        return _mypaintlib.Filler_tile_uniformity(self, is_empty, src)
    __swig_destroy__ = _mypaintlib.delete_Filler

# Register Filler in _mypaintlib:
_mypaintlib.Filler_swigregister(Filler)
class GapClosingFiller(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, max_dist, track_seep):
        _mypaintlib.GapClosingFiller_swiginit(self, _mypaintlib.new_GapClosingFiller(max_dist, track_seep))

    def fill(self, alphas, distances, dst, seeds, min_x, min_y, max_x, max_y):
        return _mypaintlib.GapClosingFiller_fill(self, alphas, distances, dst, seeds, min_x, min_y, max_x, max_y)

    def unseep(self, distances, dst, seeds, initial):
        return _mypaintlib.GapClosingFiller_unseep(self, distances, dst, seeds, initial)
    __swig_destroy__ = _mypaintlib.delete_GapClosingFiller

# Register GapClosingFiller in _mypaintlib:
_mypaintlib.GapClosingFiller_swigregister(GapClosingFiller)

def rgba_tile_from_alpha_tile(src, fill_r, fill_g, fill_b, min_x, min_y, max_x, max_y):
    return _mypaintlib.rgba_tile_from_alpha_tile(src, fill_r, fill_g, fill_b, min_x, min_y, max_x, max_y)
class DistanceBucket(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, distance):
        _mypaintlib.DistanceBucket_swiginit(self, _mypaintlib.new_DistanceBucket(distance))
    __swig_destroy__ = _mypaintlib.delete_DistanceBucket
    distance = property(_mypaintlib.DistanceBucket_distance_get)
    input = property(_mypaintlib.DistanceBucket_input_get, _mypaintlib.DistanceBucket_input_set)

# Register DistanceBucket in _mypaintlib:
_mypaintlib.DistanceBucket_swigregister(DistanceBucket)

def find_gaps(bucket, gap_output, src_mid, src_n, src_e, src_s, src_w, src_ne, src_se, src_sw, src_nw):
    return _mypaintlib.find_gaps(bucket, gap_output, src_mid, src_n, src_e, src_s, src_w, src_ne, src_se, src_sw, src_nw)
class Controller(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self):
        _mypaintlib.Controller_swiginit(self, _mypaintlib.new_Controller())

    def stop(self):
        return _mypaintlib.Controller_stop(self)

    def inc_processed(self, incr):
        return _mypaintlib.Controller_inc_processed(self, incr)

    def num_processed(self):
        return _mypaintlib.Controller_num_processed(self)

    def reset(self):
        return _mypaintlib.Controller_reset(self)
    __swig_destroy__ = _mypaintlib.delete_Controller

# Register Controller in _mypaintlib:
_mypaintlib.Controller_swigregister(Controller)

def morph(offset, morphed, tiles, strands, status_controller):
    return _mypaintlib.morph(offset, morphed, tiles, strands, status_controller)

def blur(radius, blurred, tiles, strands, status_controller):
    return _mypaintlib.blur(radius, blurred, tiles, strands, status_controller)

def get_libmypaint_brush_settings():
    return _mypaintlib.get_libmypaint_brush_settings()

def get_libmypaint_brush_inputs():
    return _mypaintlib.get_libmypaint_brush_inputs()

def gdkpixbuf_get_pixels_array(pixbuf_pyobject):
    return _mypaintlib.gdkpixbuf_get_pixels_array(pixbuf_pyobject)




# This file is part of MyPaint.
# Copyright (C) 2009-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2010-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function

from lib.gibindings import Gtk
from lib.gibindings import Gdk
import cairo

from . import windowing


"""Color history popup."""


## Module constants

popup_height = 60
bigcolor_width = popup_height
smallcolor_width = popup_height // 2


## Class definitions

class HistoryPopup (windowing.PopupWindow):
    """History popup window.

    This window is normally bound to the "X" key, and has the following
    behavior:

    1. x, x, x, draw: cycles 3 colours back;
    2. x, draw, x, draw, x, draw, ...: flips between two most recent
       colours at the time this sequence started.

    It's managed from a StateGroup: see gui.stategroup. See also
    https://github.com/mypaint/mypaint/issues/93 for discussion.

    """

    outside_popup_timeout = 0

    def __init__(self, app, model):
        super(HistoryPopup, self).__init__(app)
        # TODO: put the mouse position onto the selected color
        # FIXME: This duplicates stuff from the PopupWindow
        self.set_position(Gtk.WindowPosition.MOUSE)
        self.app = app
        self.app.kbm.add_window(self)

        hist_len = len(self.app.brush_color_manager.get_history())
        self.popup_width = bigcolor_width + (hist_len-1)*smallcolor_width

        self.set_events(Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK |
                        Gdk.EventMask.ENTER_NOTIFY_MASK |
                        Gdk.EventMask.LEAVE_NOTIFY_MASK
                        )
        self.connect("button-release-event", self.button_release_cb)
        self.connect("button-press-event", self.button_press_cb)

        self.connect("draw", self.draw_cb)

        self.set_size_request(self.popup_width, popup_height)

        # Selection index. Each enter() (i.e. keypress) advances this.
        self.selection = None

        # Reset the selection when something is drawn,
        # or when the colour history changes.
        mgr = self.app.brush_color_manager
        mgr.color_history_updated += self._reset_selection_cb
        model.canvas_area_modified += self._reset_selection_cb

    def enter(self):
        mgr = self.app.brush_color_manager
        hist = mgr.get_history()
        if self.selection is None:
            self.selection = len(hist) - 1
            color = mgr.get_color()
            if hist[self.selection] == color:
                self.selection -= 1
        else:
            self.selection = (self.selection - 1) % len(hist)

        mgr.set_color(hist[self.selection])

        # popup placement
        x, y = self.get_position()
        bigcolor_center_x = (self.selection * smallcolor_width +
                             bigcolor_width // 2)
        self.move(x + self.popup_width // 2 - bigcolor_center_x,
                  y + bigcolor_width)
        self.show_all()

        window = self.get_window()
        cursor = Gdk.Cursor.new_for_display(
            window.get_display(), Gdk.CursorType.CROSSHAIR)
        window.set_cursor(cursor)

    def leave(self, reason):
        self.hide()

    def button_press_cb(self, widget, event):
        pass

    def button_release_cb(self, widget, event):
        pass

    def _reset_selection_cb(self, *_args, **_kwargs):
        self.selection = None

    def draw_cb(self, widget, cr):
        cr.set_source_rgb(0.9, 0.9, 0.9)
        cr.paint()

        cr.set_line_join(cairo.LINE_JOIN_ROUND)

        cr.translate(0.0, popup_height/2.0)

        hist = self.app.brush_color_manager.get_history()
        for i, c in enumerate(hist):
            if i != self.selection:
                cr.scale(0.5, 0.5)

            line_width = 3.0
            distance = 2*line_width
            rect = [0, -popup_height/2.0, popup_height, popup_height]
            rect[0] += distance/2.0
            rect[1] += distance/2.0
            rect[2] -= distance
            rect[3] -= distance
            cr.rectangle(*rect)
            cr.set_source_rgb(*c.get_rgb())
            cr.fill_preserve()
            cr.set_line_width(line_width)
            cr.set_source_rgb(0, 0, 0)
            cr.stroke()
            cr.translate(popup_height, 0)

            if i != self.selection:
                cr.scale(2.0, 2.0)

        return True



# This file is part of MyPaint.
# Copyright (C) 2018-2019 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""This module implements tile-based morphological operations;
dilation, erosion and blur
"""
import logging

import lib.mypaintlib as myplib

import lib.fill_common as fc
from lib.fill_common import _FULL_TILE, _EMPTY_TILE

N = myplib.TILE_SIZE

logger = logging.getLogger(__name__)


def adjacent_tiles(tile_coord, filled):
    """ Return a tuple of tiles adjacent to the input tile coordinate.
    Adjacent tiles that are not in the tileset are replaced by the empty tile.
    """
    return tuple([filled.get(c, _EMPTY_TILE) for c in fc.adjacent(tile_coord)])


def complement_adjacent(tiles):
    """ Ensure that each tile in the input tileset has a full neighbourhood
    of eight tiles, setting missing tiles to the empty tile.

    The new set should only be used as input to tile operations, as the empty
    tile is readonly.
    """
    new = {}
    for tile_coord in tiles.keys():
        for adj_coord in fc.adjacent(tile_coord):
            if adj_coord not in tiles and adj_coord not in new:
                new[adj_coord] = _EMPTY_TILE
    tiles.update(new)


def directly_below(coord1, coord2):
    """ Return true if the first coordinate is directly below the second"""
    return coord1[0] == coord2[0] and coord1[1] == coord2[1] + 1


def strand_partition(tiles, dilating=False):
    """Partition input tiles for easier processing
    This function partitions a tile dictionary into
    two parts: one dictionary containing tiles that
    do not need to be processed further (see note),
    and list of coordinate lists, where each list
    contains vertically contiguous coordinates,
    ordered from low to high.

    note: Tiles that never need further processing are
    those that are fully opaque and with a full neighbourhood
    of identical tiles. If the "dilating" parameter is set
    to true, just being fully opaque is enough.
    :return: (final_dict, strands_list)
    """
    # Dict of coord->tile for tiles that need no further processing
    final_tiles = {}
    # Groups of contiguous tile coordinates
    strands = []
    strand = []
    previous = None
    coords = tiles.keys()
    for tile_coord in sorted(coords):
        is_full_tile = tiles[tile_coord] is _FULL_TILE
        if is_full_tile and (dilating or adj_full(tile_coord, tiles)):
            # Tile needs no processing
            final_tiles[tile_coord] = _FULL_TILE
            previous = None
            if strand:
                strands.append(strand)
                strand = []
        elif previous is None or directly_below(tile_coord, previous):
            # Either beginning of new strand, or adds to existing one
            strand.append(tile_coord)
        else:
            # Neither final, nor contiguous, begin new strand
            strands.append(strand)
            strand = [tile_coord]
        previous = tile_coord
    if strand:
        strands.append(strand)
    return final_tiles, strands


def morph(handler, offset, tiles):
    """ Either dilate or erode the given set of alpha tiles, depending
    on the sign of the offset, returning the set of morphed tiles.
    """
    # When dilating, create new tiles to account for edge overflow
    # (without checking if they are actually needed)
    if offset > 0:
        complement_adjacent(tiles)

    handler.set_stage(handler.MORPH, len(tiles))

    # Split up the coordinates of the tiles to morph, into vertically
    # contiguous strands, which can be processed more efficiently
    morphed, strands = strand_partition(tiles, offset > 0)
    # Run the morph operation (C++, conditionally threaded)
    myplib.morph(offset, morphed, tiles, strands, handler.controller)
    return morphed


def blur(handler, radius, tiles):
    """ Return the set of blurred tiles based on the input tiles.
    """
    complement_adjacent(tiles)

    handler.set_stage(handler.BLUR, len(tiles))

    blurred, strands = strand_partition(tiles, dilating=False)
    myplib.blur(radius, blurred, tiles, strands, handler.controller)
    return blurred


def adj_full(coord, tiles):
    return all(t is _FULL_TILE for t in adjacent_tiles(coord, tiles))



# This file is part of MyPaint.
# Copyright (C) 2016-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Exposes information about the settings and inputs libmypaint uses."""


# Imports:

from __future__ import division, print_function

import lib.mypaintlib as _mypaintlib


# Public variables:

#: List of info about all brush engine inputs known to libmypaint.
#: Elements are `BrushInputInfo` objects.
inputs = []

#: Contents of `inputs`, organized by internal name.
inputs_dict = {}

#: List of info about all brush settings known to libmypaint.
#: Elements are `BrushSettingInfo` objects.
settings = []

#: Contents of `settings`, organized by internal name.
settings_dict = {}

#: List of only the settings that should be visible in a brush editor UI.
settings_visible = []

#: Migration plan and calculations for old setting names.
settings_migrate = {
    # old cname              new cname        scale function
    'color_hue': ('change_color_h', lambda y: y * 64.0 / 360.0),
    'color_saturation': ('change_color_hsv_s', lambda y: y * 128.0 / 256.0),
    'color_value': ('change_color_v', lambda y: y * 128.0 / 256.0),
    'speed_slowness': ('speed1_slowness', None),
    'change_color_s': ('change_color_hsv_s', None),
    'stroke_treshold': ('stroke_threshold', None),
}


# Class definitions:

class BrushInputInfo:
    """Information about an input known by the libmypaint brush engine.

    Inputs are reflections of the data your tablet pen sends, and
    include things like tilt, pressure, and position. There are special
    inputs too, such as the speed and direction of drawing, a random
    input, and one which can be specified by the user to contain the
    effect of one or more settings.

    >>> for bi in inputs:
    ...     assert 0 <= bi.index < len(inputs)
    ...     assert bi.name in inputs_dict
    ...     for f in BrushInputInfo._FIELDS:
    ...         assert getattr(bi, f) is not None
    ...     assert bi.hard_max > bi.hard_min
    ...     assert bi.soft_max > bi.soft_min
    ...     assert bi.hard_max >= bi.soft_max
    ...     assert bi.hard_min <= bi.soft_min

    :ivar str name: Internal setting name.
    :ivar float hard_min:
    :ivar float soft_min:
    :ivar float normal:
    :ivar float hard_max:
    :ivar float soft_max:
    :ivar unicode dname: Localized human-readable name.
    :ivar unicode tooltip: Localized short technical description.
    :ivar int index: Index of the object in `inputs`.

    """

    _FIELDS = (
        "name",
        "hard_min", "soft_min",
        "normal",
        "hard_max", "soft_max",
        "dname", "tooltip",
        "index",
    )

    def __init__(self, **kwargs):
        for k in self._FIELDS:
            setattr(self, k, kwargs.get(k))

    def __repr__(self):
        return u"<{classname} {fields}>".format(
            classname = self.__class__.__name__,
            fields = {k: getattr(self, k, None) for k in self._FIELDS},
        )


class BrushSettingInfo:
    """Representation of a brush setting known to libmypaint.

    Settings are what the user tweaks in the brush editor. They include
    things like the radius or the smudge factor. Think of them as
    user-defined functions that use the current value of zero or more
    BrushInputs to form a single output value.

    >>> for bs in settings:
    ...      assert 0 <= bs.index < len(settings)
    ...      assert bs.cname in settings_dict
    ...      for f in BrushSettingInfo._FIELDS:
    ...          assert getattr(bs, f) is not None
    ...      assert bs.max > bs.min

    :ivar str cname: Internal name.
    :ivar str name: Localized, short, human-readable name.
    :ivar bool constant:
    :ivar float min:
    :ivar float default:
    :ivar float max:
    :ivar unicode tooltip: Localized, human-readable technical description.
    :ivar int index: Index of the object in `settings`.

    """

    _FIELDS = [
        "cname",
        "name",
        "constant",
        "min", "default", "max",
        "tooltip",
        "index",
    ]

    def __init__(self, **kwargs):
        for k in self._FIELDS:
            setattr(self, k, kwargs.get(k))


# Module initialization:

for i, info_dict in enumerate(_mypaintlib.get_libmypaint_brush_inputs()):
    info_dict["index"] = i
    input = BrushInputInfo(**info_dict)
    inputs.append(input)
    inputs_dict[input.name] = input

for i, info_dict in enumerate(_mypaintlib.get_libmypaint_brush_settings()):
    info_dict["index"] = i
    setting = BrushSettingInfo(**info_dict)
    settings.append(setting)
    settings_dict[setting.cname] = setting

_settings_hidden = set(["color_h", "color_s", "color_v"])
for s in settings:
    if s.cname in _settings_hidden:
        continue
    settings_visible.append(s)


# Module testing:

def _test():
    import doctest
    doctest.testmod()


if __name__ == '__main__':
    _test()
